import tensorflow as tf

from tensorflow_similarity.types import Tensor
from tensorflow_similarity.augmenters.augmentation_utils.random_apply import (
    random_apply,
)


def random_random_flip_left_right(
    image: Tensor, p: float = 0.5
) -> Tensor:
    def _transform(image: Tensor) -> Tensor:
        return tf.image.random_flip_left_right(image)

    return random_apply(_transform, p=p, x=image)


def random_random_flip_top_bottom(
    image: Tensor, p: float = 0.5
) -> Tensor:
    def _transform(image: Tensor) -> Tensor:
        return tf.image.random_flip_up_down(image)

    return random_apply(_transform, p=p, x=image)



import re

import pytest

from tensorflow_similarity.architectures import resnet18


def test_include_top():
    input_shape = (32, 32, 3)
    resnet = resnet18.ResNet18Sim(input_shape, include_top=True)

    # The second to last layer should use gem pooling when include_top is True
    assert resnet.layers[-2].name == 'gem_pool'
    assert resnet.layers[-2].p == 3.0
    # The default is l2_norm True, so we expect the last layer to be
    # MetricEmbedding.
    assert re.match('metric_embedding', resnet.layers[-1].name) is not None


def test_l2_norm_false():
    input_shape = (32, 32, 3)
    resnet = resnet18.ResNet18Sim(
            input_shape,
            include_top=True,
            l2_norm=False)

    # The second to last layer should use gem pooling when include_top is True
    assert resnet.layers[-2].name == 'gem_pool'
    assert resnet.layers[-2].p == 3.0
    # If l2_norm is False, we should return a dense layer as the last layer.
    assert re.match('dense', resnet.layers[-1].name) is not None


@pytest.mark.parametrize(
    "pooling, name",
    zip(['gem', 'avg', 'max'], ['gem_pool', 'avg_pool', 'max_pool']),
    ids=['gem', 'avg', 'max']
)
def test_include_top_false(pooling, name):
    input_shape = (32, 32, 3)
    resnet = resnet18.ResNet18Sim(
            input_shape,
            include_top=False,
            pooling=pooling)

    # The second to last layer should use gem pooling when include_top is True
    assert resnet.layers[-1].name == name



import re

import pytest
import tensorflow as tf

from tensorflow_similarity.architectures import resnet50


def test_build_full():
    input_layer = tf.keras.layers.Input((224, 224, 3))
    output = resnet50.build_resnet(input_layer, "imagenet", "full")

    effnet = output._keras_history.layer

    assert effnet.trainable

    total_layer_count = 0
    trainable_layer_count = 0
    for layer in effnet._self_tracked_trackables:
        total_layer_count += 1
        if layer.trainable:
            trainable_layer_count += 1

    assert total_layer_count == 175
    assert trainable_layer_count == 122


def test_build_frozen():
    input_layer = tf.keras.layers.Input((240, 240, 3))
    output = resnet50.build_resnet(input_layer, "imagenet", "frozen")

    effnet = output._keras_history.layer

    assert not effnet.trainable

    total_layer_count = 0
    trainable_layer_count = 0
    for layer in effnet._self_tracked_trackables:
        total_layer_count += 1
        if layer.trainable:
            trainable_layer_count += 1

    assert total_layer_count == 175
    assert trainable_layer_count == 0


def test_build_partial():
    input_layer = tf.keras.layers.Input((224, 224, 3))
    output = resnet50.build_resnet(input_layer, "imagenet", "partial")

    effnet = output._keras_history.layer

    assert effnet.trainable

    total_layer_count = 0
    trainable_layer_count = 0
    excluded_layers = 0
    for layer in effnet._self_tracked_trackables:
        total_layer_count += 1
        if layer.trainable:
            trainable_layer_count += 1
            # Check if any of the excluded layers are trainable
            if not re.search("^conv5|^top", layer.name):
                excluded_layers += 1
            if isinstance(layer, tf.keras.layers.BatchNormalization):
                excluded_layers += 1

    assert total_layer_count == 175
    assert trainable_layer_count == 22
    assert excluded_layers == 0


def test_build_unsupported_trainable():
    input_layer = tf.keras.layers.Input((224, 224, 3))
    msg = "foo is not a supported option for 'trainable'."
    with pytest.raises(ValueError, match=msg):
        _ = resnet50.build_resnet(input_layer, "imagenet", "foo")


def test_include_top():
    input_shape = (224, 224, 3)
    resnet = resnet50.ResNet50Sim(input_shape, include_top=True)

    # The second to last layer should use gem pooling when include_top is True
    assert resnet.layers[-2].name == 'gem_pool'
    assert resnet.layers[-2].p == 3.0
    # The default is l2_norm True, so we expect the last layer to be
    # MetricEmbedding.
    assert re.match('metric_embedding', resnet.layers[-1].name) is not None


def test_l2_norm_false():
    input_shape = (224, 224, 3)
    resnet = resnet50.ResNet50Sim(
            input_shape,
            include_top=True,
            l2_norm=False)

    # The second to last layer should use gem pooling when include_top is True
    assert resnet.layers[-2].name == 'gem_pool'
    assert resnet.layers[-2].p == 3.0
    # If l2_norm is False, we should return a dense layer as the last layer.
    assert re.match('dense', resnet.layers[-1].name) is not None


@pytest.mark.parametrize(
    "pooling, name",
    zip(['gem', 'avg', 'max'], ['gem_pool', 'avg_pool', 'max_pool']),
    ids=['gem', 'avg', 'max']
)
def test_include_top_false(pooling, name):
    input_shape = (224, 224, 3)
    resnet = resnet50.ResNet50Sim(
            input_shape,
            include_top=False,
            pooling=pooling)

    # The second to last layer should use gem pooling when include_top is True
    assert resnet.layers[-1].name == name



# Copyright 2021 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from typing import Callable, List, Optional

import tensorflow as tf

from tensorflow_similarity.augmenters import Augmenter
from tensorflow_similarity.types import Tensor


class ContrastiveAugmenter(Augmenter):
    def __init__(
        self, process: Callable, num_cpu: Optional[int] = os.cpu_count()
    ):
        self.process = process
        self.num_cpu = num_cpu

    def augment(
        self, x: Tensor, y: Tensor, num_views: int, is_warmup: bool
    ) -> List[Tensor]:
        with tf.device("/cpu:0"):
            inputs = tf.stack(x)

            views = []
            for _ in range(num_views):
                # multi-cor augementations
                view = tf.map_fn(
                    self.process, inputs, parallel_iterations=self.num_cpu
                )
                views.append(view)
            return views



import re

import pytest
import tensorflow as tf

from tensorflow_similarity.architectures import efficientnet

# TODO(ovallis): rewrite these tests so they aren't so brittle.
TF_MAJOR_VERSION = int(tf.__version__.split(".")[0])
TF_MINOR_VERSION = int(tf.__version__.split(".")[1])


def tf_version_check(major_version, minor_version):
    if TF_MAJOR_VERSION <= major_version and TF_MINOR_VERSION < minor_version:
        return True

    return False


def test_build_effnet_b0_full():
    input_layer = tf.keras.layers.Input((224, 224, 3))
    output = efficientnet.build_effnet(input_layer, "b0", "imagenet", "full")

    effnet = output._keras_history.layer

    assert effnet.name == "efficientnetb0"
    assert effnet.trainable

    total_layer_count = 0
    trainable_layer_count = 0
    for layer in effnet._self_tracked_trackables:
        total_layer_count += 1
        if layer.trainable:
            trainable_layer_count += 1

    if tf_version_check(2, 9):
        expected_total_layer_count = 237
        expected_trainable_layer_count = 188
    else:
        expected_total_layer_count = 238
        expected_trainable_layer_count = 189

    assert total_layer_count == expected_total_layer_count
    assert trainable_layer_count == expected_trainable_layer_count


def test_build_effnet_b1_frozen():
    input_layer = tf.keras.layers.Input((240, 240, 3))
    output = efficientnet.build_effnet(input_layer, "b1", "imagenet", "frozen")

    effnet = output._keras_history.layer

    assert effnet.name == "efficientnetb1"
    assert not effnet.trainable

    total_layer_count = 0
    trainable_layer_count = 0
    for layer in effnet._self_tracked_trackables:
        total_layer_count += 1
        if layer.trainable:
            trainable_layer_count += 1

    if tf_version_check(2, 9):
        expected_total_layer_count = 339
    else:
        expected_total_layer_count = 340

    assert total_layer_count == expected_total_layer_count
    assert trainable_layer_count == 0


def test_build_effnet_b0_partial():
    input_layer = tf.keras.layers.Input((224, 224, 3))
    output = efficientnet.build_effnet(input_layer, "b0", "imagenet", "partial")

    effnet = output._keras_history.layer

    assert effnet.name == "efficientnetb0"
    assert effnet.trainable

    total_layer_count = 0
    trainable_layer_count = 0
    excluded_layers = 0
    for layer in effnet._self_tracked_trackables:
        total_layer_count += 1
        if layer.trainable:
            trainable_layer_count += 1
            # Check if any of the excluded layers are trainable
            if not re.search("^block[5,6,7]|^top", layer.name):
                excluded_layers += 1
            if isinstance(layer, tf.keras.layers.BatchNormalization):
                excluded_layers += 1

    expected_trainable_layer_count = 93

    if tf_version_check(2, 9):
        expected_total_layer_count = 237
    else:
        expected_total_layer_count = 238

    assert total_layer_count == expected_total_layer_count
    assert trainable_layer_count == expected_trainable_layer_count
    assert excluded_layers == 0


def test_build_effnet_unsupported_trainable():
    input_layer = tf.keras.layers.Input((224, 224, 3))
    msg = "foo is not a supported option for 'trainable'."
    with pytest.raises(ValueError, match=msg):
        _ = efficientnet.build_effnet(input_layer, "b0", "imagenet", "foo")


def test_unsuported_varient():
    input_shape = (224, 224, 3)
    msg = "Unknown efficientnet variant. Valid B0...B7"
    with pytest.raises(ValueError, match=msg):
        _ = efficientnet.EfficientNetSim(input_shape, 128, "bad_varient")


def test_include_top():
    input_shape = (224, 224, 3)
    effnet = efficientnet.EfficientNetSim(input_shape, include_top=True)

    # The second to last layer should use gem pooling when include_top is True
    assert effnet.layers[-2].name == "gem_pool"
    assert effnet.layers[-2].p == 3.0
    # The default is l2_norm True, so we expect the last layer to be
    # MetricEmbedding.
    assert re.match("metric_embedding", effnet.layers[-1].name) is not None


def test_l2_norm_false():
    input_shape = (224, 224, 3)
    effnet = efficientnet.EfficientNetSim(
        input_shape, include_top=True, l2_norm=False
    )

    # The second to last layer should use gem pooling when include_top is True
    assert effnet.layers[-2].name == "gem_pool"
    assert effnet.layers[-2].p == 3.0
    # If l2_norm is False, we should return a dense layer as the last layer.
    assert re.match("dense", effnet.layers[-1].name) is not None


@pytest.mark.parametrize(
    "pooling, name",
    zip(["gem", "avg", "max"], ["gem_pool", "avg_pool", "max_pool"]),
    ids=["gem", "avg", "max"],
)
def test_include_top_false(pooling, name):
    input_shape = (224, 224, 3)
    effnet = efficientnet.EfficientNetSim(
        input_shape, include_top=False, pooling=pooling
    )

    # The second to last layer should use gem pooling when include_top is True
    assert effnet.layers[-1].name == name



from unittest import TestCase

import numpy as np
import pytest
import tensorflow as tf

from tensorflow_similarity.schedules import WarmUpCosine

testdata = [
    range(10),
    [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9],
    np.arange(10),
    tf.range(10, dtype="float32"),
]

test_ids = ["python range", "python list", "numpy arange", "tf range"]


@pytest.mark.parametrize("steps", testdata, ids=test_ids)
def test_warmup_cosine(steps):
    warmup_cosine = WarmUpCosine(
        initial_learning_rate=1.0,
        decay_steps=10,
        warmup_steps=2,
        warmup_learning_rate=0.5,
        alpha=0.1,
        name="foo",
    )

    lrs = warmup_cosine(steps)

    expected_lrs = tf.constant(
        [
            0.5,
            0.7334816,
            0.9140576,
            0.8145034,
            0.68905765,
            0.54999995,
            0.41094226,
            0.28549662,
            0.18594232,
            0.12202458,
        ]
    )

    np.testing.assert_allclose(expected_lrs, lrs, rtol=1e-06)


def test_warmup_cosine_no_warmup():
    warmup_cosine = WarmUpCosine(
        initial_learning_rate=1.0,
        decay_steps=10,
        warmup_steps=0,
        warmup_learning_rate=0.5,
        alpha=0.1,
        name="foo",
    )

    lrs = warmup_cosine(range(10))

    expected_lrs = tf.constant(
        [
            1.0,
            0.977975,
            0.9140576,
            0.8145034,
            0.68905765,
            0.54999995,
            0.41094226,
            0.28549662,
            0.18594232,
            0.12202458,
        ]
    )

    np.testing.assert_allclose(expected_lrs, lrs, rtol=1e-06)


def test_warmup_cosine_config():
    warmup_cosine = WarmUpCosine(
        initial_learning_rate=1.0,
        decay_steps=10,
        warmup_steps=2,
        warmup_learning_rate=0.0,
        alpha=0.1,
        name="foo",
    )

    config = warmup_cosine.get_config()

    expected_config = {
        "initial_learning_rate": 1.0,
        "decay_steps": 10,
        "alpha": 0.1,
        "warmup_learning_rate": 0.0,
        "warmup_steps": 2,
        "name": "foo",
    }

    TestCase().assertDictEqual(expected_config, config)


def test_warmup_cosine_assert_smaller_warmup_learning_rate():
    msg = "warmup_learning_rate must be smaller than the initial_learning_rate"
    with pytest.raises(ValueError, match=msg):
        _ = WarmUpCosine(
            initial_learning_rate=1.0,
            decay_steps=10,
            warmup_steps=5,
            warmup_learning_rate=1.5,
        )


def test_warmup_cosine_assert_smaller_warmup_steps():
    msg = "warmup_steps must be smaller than the decay_steps"
    with pytest.raises(ValueError, match=msg):
        _ = WarmUpCosine(
            initial_learning_rate=1.0,
            decay_steps=10,
            warmup_steps=20,
        )



# Copyright 2020 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""SimCLR Loss.
    A Simple Framework for Contrastive Learning of Visual Representations
    https://arxiv.org/abs/2002.05709
"""
from typing import Any, Callable, Dict, Optional

import tensorflow as tf

from tensorflow_similarity.types import FloatTensor

LARGE_NUM = 1e9


@tf.keras.utils.register_keras_serializable(package="Similarity")
class SimCLRLoss(tf.keras.losses.Loss):
    """SimCLR Loss.
    [A Simple Framework for Contrastive Learning of Visual Representations](https://arxiv.org/abs/2002.05709)
    code adapted from [orignal github](https://github.com/google-research/simclr/tree/master/tf2)
    """

    def __init__(
        self,
        temperature: float = 0.05,
        margin: float = 0.001,
        reduction: Callable = tf.keras.losses.Reduction.AUTO,
        name: Optional[str] = None,
        **kwargs
    ):
        super().__init__(reduction=reduction, name=name, **kwargs)
        self.temperature = temperature
        self.margin = margin

    def call(self, za: FloatTensor, zb: FloatTensor) -> FloatTensor:
        """Compute the lost.

        Args:
            za: Embedding A
            zb: Embedding B

        Returns:
            loss
        """
        # compute the diagonal
        batch_size = tf.shape(za)[0]
        diag = tf.one_hot(tf.range(batch_size), batch_size)

        # We expect za and zb to be rank 2 tensors.
        za = tf.math.l2_normalize(za, axis=1)
        zb = tf.math.l2_normalize(zb, axis=1)

        # compute pairwise
        ab = tf.matmul(za, zb, transpose_b=True)
        aa = tf.matmul(za, za, transpose_b=True)

        # divide by temperature once to go faster
        ab = ab / self.temperature
        aa = aa / self.temperature

        # set the diagonal to a large negative number to ensure that z_i * z_i
        # is close to zero in the cross entropy.
        aa = aa - diag * LARGE_NUM

        distances = tf.concat((ab, aa), axis=1)
        labels = tf.one_hot(tf.range(batch_size), batch_size * 2)

        # 1D tensor
        per_example_loss = tf.nn.softmax_cross_entropy_with_logits(
            labels, distances
        )

        # 1D tensor
        loss: FloatTensor = per_example_loss * 0.5 + self.margin

        return loss

    def get_config(self) -> Dict[str, Any]:
        config = {
            "temperature": self.temperature,
            "margin": self.margin,
        }
        base_config = super().get_config()
        return {**base_config, **config}



from typing import Tuple

import tensorflow as tf

from tensorflow_similarity.types import Tensor
from tensorflow_similarity.augmenters.augmentation_utils.random_apply import (
    random_apply,
)


def _compute_crop_shape(
    image_height: int,
    image_width: int,
    aspect_ratio: float,
    crop_proportion: float,
) -> Tuple[int, int]:
    """Compute aspect ratio-preserving shape for central crop.

    The resulting shape retains `crop_proportion` along one side and a
    proportion less than or equal to `crop_proportion` along the other side.

    Args:
      image_height: Height of image to be cropped.
      image_width: Width of image to be cropped.
      aspect_ratio: Desired aspect ratio (width / height) of output.
      crop_proportion: Proportion of image to retain along the less-cropped
      side.

    Returns:
      crop_height: Height of image after cropping.
      crop_width: Width of image after cropping.
    """
    image_width_float = tf.cast(image_width, tf.float32)
    image_height_float = tf.cast(image_height, tf.float32)

    def _requested_aspect_ratio_wider_than_image():
        crop_height = tf.cast(
            tf.math.rint(crop_proportion / aspect_ratio * image_width_float),
            tf.int32,
        )
        crop_width = tf.cast(
            tf.math.rint(crop_proportion * image_width_float), tf.int32
        )
        return crop_height, crop_width

    def _image_wider_than_requested_aspect_ratio():
        crop_height = tf.cast(
            tf.math.rint(crop_proportion * image_height_float), tf.int32
        )
        crop_width = tf.cast(
            tf.math.rint(crop_proportion * aspect_ratio * image_height_float),
            tf.int32,
        )
        return crop_height, crop_width

    crop_height, crop_width = tf.cond(
        aspect_ratio > image_width_float / image_height_float,
        _requested_aspect_ratio_wider_than_image,
        _image_wider_than_requested_aspect_ratio,
    )

    return crop_height, crop_width


def center_crop(
    image: Tensor, height: int, width: int, crop_proportion: float
) -> Tensor:
    """Crops to center of image and rescales to desired size.

    Args:
      image: Image Tensor to crop.
      height: Height of image to be cropped.
      width: Width of image to be cropped.
      crop_proportion: Proportion of image to retain along the less-cropped side.

    Returns:
      A `height` x `width` x channels Tensor holding a central crop of `image`.
    """
    shape = tf.shape(image)
    image_height = shape[0]
    image_width = shape[1]
    crop_height, crop_width = _compute_crop_shape(
        image_height, image_width, height / width, crop_proportion
    )
    offset_height = ((image_height - crop_height) + 1) // 2
    offset_width = ((image_width - crop_width) + 1) // 2
    image = tf.image.crop_to_bounding_box(
        image, offset_height, offset_width, crop_height, crop_width
    )

    image = tf.image.resize(
        [image], [height, width], method=tf.image.ResizeMethod.BICUBIC
    )[0]

    return image


def distorted_bounding_box_crop(
    image: Tensor,
    bbox: Tensor,
    min_object_covered: float = 0.1,
    aspect_ratio_range: Tuple[float, float] = (0.75, 1.33),
    area_range: Tuple[float, float] = (0.05, 1.0),
    max_attempts: int = 100,
    scope: bool = None,
) -> Tensor:
    """Generates cropped_image using one of the bboxes randomly distorted.

    See `tf.image.sample_distorted_bounding_box` for more documentation.

    Args:
      image: `Tensor` of image data.
      bbox: `Tensor` of bounding boxes arranged `[1, num_boxes, coords]`
          where each coordinate is [0, 1) and the coordinates are arranged
          as `[ymin, xmin, ymax, xmax]`. If num_boxes is 0 then use the whole
          image.
      min_object_covered: An optional `float`. Defaults to `0.1`. The cropped
          area of the image must contain at least this fraction of any bounding
          box supplied.
      aspect_ratio_range: An optional list of `float`s. The cropped area of the
          image must have an aspect ratio = width / height within this range.
      area_range: An optional list of `float`s. The cropped area of the image
          must contain a fraction of the supplied image within in this range.
      max_attempts: An optional `int`. Number of attempts at generating a cropped
          region of the image of the specified constraints. After `max_attempts`
          failures, return the entire image.
      scope: Optional `str` for name scope.
    Returns:
      cropped image `Tensor`.
    """
    with tf.name_scope(scope or "distorted_bounding_box_crop"):
        shape = tf.shape(image)
        sample_distorted_bounding_box = tf.image.sample_distorted_bounding_box(
            shape,
            bounding_boxes=bbox,
            min_object_covered=min_object_covered,
            aspect_ratio_range=aspect_ratio_range,
            area_range=area_range,
            max_attempts=max_attempts,
            use_image_if_no_bounding_boxes=True,
        )
        bbox_begin, bbox_size, _ = sample_distorted_bounding_box

        # Crop the image to the specified bounding box.
        offset_y, offset_x, _ = tf.unstack(bbox_begin)
        target_height, target_width, _ = tf.unstack(bbox_size)
        image = tf.image.crop_to_bounding_box(
            image, offset_y, offset_x, target_height, target_width
        )

        return image


def crop_and_resize(
    image: Tensor,
    height: int,
    width: int,
    area_range: Tuple[float, float] = (0.08, 1.0),
) -> Tensor:
    """Make a random crop and resize it to height `height` and width `width`.

    Args:
      image: Tensor representing the image.
      height: Desired image height.
      width: Desired image width.
      area_range: An optional list of `float`s. The cropped area of the image
          must contain a fraction of the supplied image within in this range.

    Returns:
      A `height` x `width` x channels Tensor holding a random crop of `image`.
    """
    bbox = tf.constant([0.0, 0.0, 1.0, 1.0], dtype=tf.float32, shape=[1, 1, 4])
    aspect_ratio = width / height
    image = distorted_bounding_box_crop(
        image,
        bbox,
        min_object_covered=0.1,
        aspect_ratio_range=(3.0 / 4 * aspect_ratio, 4.0 / 3.0 * aspect_ratio),
        area_range=area_range,
        max_attempts=100,
        scope=None,
    )
    img = tf.image.resize(
        [image], [height, width], method=tf.image.ResizeMethod.BICUBIC
    )[0]
    return img


def crop_random_resized_crop(
    image: Tensor,
    height: int,
    width: int,
    min_cropsize_multiplier: float = 0.75,
    max_cropsize_multiplier: float = 1,
):
    """
    Crops and resizes image by getting a random square of size ranging from 75% of the width to 100% of width.
    This cropped portion is then resized to the original height and width, which may cause stretching and
    distortion for rectangular images.
    """

    rand_width = tf.random.uniform(
        shape=[],
        minval=int(min_cropsize_multiplier * width),
        maxval=int(max_cropsize_multiplier * width),
        dtype=tf.int32,
    )

    crop = tf.image.random_crop(image, (rand_width, rand_width, 3))
    crop_resize = tf.image.resize(crop, (height, width))
    return crop_resize


def random_resized_crop(
    # this utility function is supposed to match
    # torchvision.transforms.RandomResizedCrop
    # Only this will work with Barlow Twins and VicReg,
    # everything else will cause significant performance
    # losses
    image: Tensor,
    height: int,
    width: int,
    min_cropsize_multiplier: float = 0.75,
    max_cropsize_multiplier: float = 1,
    p: float = 1.0,
):
    def _transform(
        image: Tensor,
    ) -> Tensor:  # pylint: disable=missing-docstring
        image = crop_random_resized_crop(
            image,
            height,
            width,
            min_cropsize_multiplier,
            max_cropsize_multiplier,
        )
        return image

    return random_apply(_transform, p=p, x=image)


def random_crop_with_resize(
    image: Tensor, height: int, width: int, p: float = 1.0
) -> Tensor:
    """Randomly crop and resize an image.

    Args:
      image: `Tensor` representing an image of arbitrary size.
      height: Height of output image.
      width: Width of output image.
      p: Probability of applying this transformation.

    Returns:
      A preprocessed image `Tensor`.
    """

    def _transform(
        image: Tensor,
    ) -> Tensor:  # pylint: disable=missing-docstring
        image = crop_and_resize(image, height, width)
        return image

    return random_apply(_transform, p=p, x=image)



# Copyright 2021 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any, Dict

import tensorflow as tf

from tensorflow_similarity.types import FloatTensor


@tf.keras.utils.register_keras_serializable(package="Similarity")
class WarmUpCosine(tf.keras.optimizers.schedules.LearningRateSchedule):
    """A LearningRateSchedule that uses a cosine decay schedule with a warmup period.

    This learning rate schedule is useful for training when using the Barlow Twin Loss.

    The warmup period applies a linear scaling to the CosineDecay schedule.
    """

    def __init__(
        self,
        initial_learning_rate: float,
        decay_steps: int,
        warmup_steps: int,
        warmup_learning_rate: float = 0.0,
        alpha: float = 0.0,
        name: str = "WarmUpCosine",
    ):
        """Applies cosine decay to the learning rate.

        Args:
          initial_learning_rate: A scalar `float32` or `float64` Tensor or a
            Python number. The initial learning rate.
          decay_steps: A scalar `int32` or `int64` `Tensor` or a Python number.
            Number of steps to decay over.
          warmup_steps: A scalar `int32` or `int64` `Tensor` or a Python number.
            Number of steps to warmup over. Must be smaller than the number of
            decay_steps.
          warmup_learning_rate: A scalar `float32` or `float64` Tensor or a
            Python number. The initial warmup learning rate. Must be smaller than
            the initial_learning_rate. Defaults to 0.0.
          alpha: A scalar `float32` or `float64` Tensor or a Python number.
            Minimum learning rate value as a fraction of initial_learning_rate.
            Defaults to 0.0.
          name: String. Optional name of the operation. Defaults to 'WarmUpCosine'.
        """

        super().__init__()

        if warmup_learning_rate > initial_learning_rate:
            raise ValueError(
                "warmup_learning_rate must be smaller than the initial_learning_rate"
            )

        if warmup_steps > decay_steps:
            raise ValueError(
                "warmup_steps must be smaller than the decay_steps"
            )
        self.initial_learning_rate = initial_learning_rate
        self.decay_steps = decay_steps
        self.alpha = alpha
        self.warmup_learning_rate = warmup_learning_rate
        self.warmup_steps = warmup_steps
        self.name = name

        self.cosine_decay = tf.keras.optimizers.schedules.CosineDecay(
            initial_learning_rate=initial_learning_rate,
            decay_steps=decay_steps,
            alpha=alpha,
        )
        # Compute the warmup increment.
        self.tf_initial_learning_rate = tf.convert_to_tensor(
            self.initial_learning_rate, name="initial_learning_rate"
        )
        self.dtype = self.tf_initial_learning_rate.dtype
        self.learning_rate_delta = tf.convert_to_tensor(
            self.warmup_learning_rate / self.initial_learning_rate, self.dtype
        )
        self.warmup_inc = tf.math.divide_no_nan(
            (1.0 - self.learning_rate_delta),
            tf.convert_to_tensor(self.warmup_steps, self.dtype),
        )

        # If the warmup increment is zero we have no warm up phase and we set
        # the learning rate delta to 1.0 to ensure the warmup_scaler value is
        # always fixed at 1.0.
        if self.warmup_inc == 0:
            self.learning_rate_delta = tf.constant([1.0], self.dtype)

    def __call__(self, step: FloatTensor) -> FloatTensor:
        global_step_recomp = tf.cast(step, self.dtype)
        warmup_scaler = tf.minimum(
            1.0, self.warmup_inc * global_step_recomp + self.learning_rate_delta
        )
        learning_rate: FloatTensor = (
            self.cosine_decay(global_step_recomp) * warmup_scaler
        )
        return learning_rate

    def get_config(self) -> Dict[str, Any]:
        return {
            "initial_learning_rate": self.initial_learning_rate,
            "decay_steps": self.decay_steps,
            "alpha": self.alpha,
            "warmup_learning_rate": self.warmup_learning_rate,
            "warmup_steps": self.warmup_steps,
            "name": self.name,
        }



import pytest
import numpy as np
import tensorflow as tf
from tensorflow_similarity.distances import (
    CosineDistance,
    InnerProductSimilarity,
)
from tensorflow_similarity.distances import EuclideanDistance
from tensorflow_similarity.distances import ManhattanDistance
from tensorflow_similarity.distances import SNRDistance
from tensorflow_similarity.distances import distance_canonicalizer
from tensorflow_similarity.distances import DISTANCES


def test_distance_mapping():
    for d in DISTANCES:

        # self naming
        d2 = distance_canonicalizer(d.name)
        assert d2.name == d.name

        # aliases
        for a in d.aliases:
            d2 = distance_canonicalizer(a)
            assert d2.name == d.name


def test_distance_passthrough():
    "Canonilizer is expected to return distance object as is"
    d = EuclideanDistance()
    d2 = distance_canonicalizer(d)
    assert d == d2


def test_non_existing_distance():
    with pytest.raises(ValueError):
        distance_canonicalizer("notadistance")


def test_inner_product_similarity():
    # pairwise
    a = tf.convert_to_tensor([[1.0, 1.0, 1.0], [1.0, 1.0, 1.0]])
    d = InnerProductSimilarity()
    vals = d(a, a)
    assert tf.math.reduce_all(tf.shape(vals) == (2, 2))
    assert tf.round(tf.reduce_sum(vals)) == 12


def test_inner_product_key():
    a = tf.convert_to_tensor([
        [0.0, 0.0],
        [0.0, 1.0],
        [1.0, 1.0],
        [3.0, 0.0]
    ])
    b = tf.convert_to_tensor([
        [0.0, 1.0],
        [1.0, 1.0],
        [3.0, 0.0]
    ])
    d = InnerProductSimilarity()
    vals = d(a, b)
    expected = tf.constant([
        [0., 0., 0.],
        [1., 1., 0.],
        [1., 2., 3.],
        [0., 3., 9.]
    ])
    tf.assert_equal(vals, expected)


def test_inner_product_opposite():
    a = tf.convert_to_tensor([[0.0, 1.0], [1.0, 0.0]])
    d = InnerProductSimilarity()
    vals = d(a, a)
    assert tf.math.reduce_all(tf.shape(vals) == (2, 2))
    assert tf.round(tf.reduce_sum(vals)) == 2


def test_inner_product_vals():
    a = tf.nn.l2_normalize([[0.1, 0.3, 0.2], [0.0, 0.1, 0.5]], axis=-1)
    d = InnerProductSimilarity()
    vals = d(a, a)
    assert tf.math.reduce_all(tf.shape(vals) == (2, 2))
    assert vals[0][0] == 1
    assert vals[0][1] == 0.68138516


def test_cosine_key():
    a = tf.convert_to_tensor([
        [0.0, 0.0],
        [0.0, 1.0],
        [1.0, 1.0],
        [3.0, 0.0]
    ])
    b = tf.convert_to_tensor([
        [0.0, 1.0],
        [1.0, 1.0],
        [3.0, 0.0]
    ])
    d = CosineDistance()
    vals = d(a, b)
    expected = tf.constant([
        [1., 1., 1.],
        [0., 0., 1.],
        [0., 0., 0.],
        [1., 0., 0.]
    ])
    tf.assert_equal(vals, expected)


def test_cosine_same():
    # pairwise
    a = tf.convert_to_tensor([[1.0, 1.0, 1.0], [1.0, 1.0, 1.0]])
    d = CosineDistance()
    vals = d(a, a)
    assert tf.math.reduce_all(tf.shape(vals) == (2, 2))
    assert tf.round(tf.reduce_sum(vals)) == 0


def test_cosine_opposite():
    a = tf.convert_to_tensor([[0.0, 1.0], [1.0, 0.0]])
    d = CosineDistance()
    vals = d(a, a)
    assert tf.math.reduce_all(tf.shape(vals) == (2, 2))
    assert tf.round(tf.reduce_sum(vals)) == 2


def test_cosine_vals():
    a = tf.nn.l2_normalize([[0.1, 0.3, 0.2], [0.0, 0.1, 0.5]], axis=-1)
    d = CosineDistance()
    vals = d(a, a)
    assert vals[0][0] == 0
    assert vals[0][1] == 0.31861484


def test_euclidean():
    a = tf.convert_to_tensor([[0.0, 3.0], [4.0, 0.0]])
    d = EuclideanDistance()
    vals = d(a, a)
    assert tf.math.reduce_all(tf.shape(vals) == (2, 2))
    assert tf.round(tf.reduce_sum(vals)) == 10


def test_euclidean_key():
    a = tf.convert_to_tensor([
        [0.0, 1.0],
        [0.0, 1.0],
        [1.0, 1.0],
    ])
    b = tf.convert_to_tensor([
        [2.0, 1.0],
        [1.0, 1.0],
    ])
    d = EuclideanDistance()
    vals = d(a, b)
    expected = tf.constant([
        [2., 1.],
        [2., 1.],
        [1., 0.]
    ])
    tf.assert_equal(vals, expected)


def test_euclidean_same():
    a = tf.convert_to_tensor([[1.0, 1.0], [1.0, 1.0]])
    d = EuclideanDistance()
    vals = d(a, a)
    assert tf.math.reduce_all(tf.shape(vals) == (2, 2))
    assert tf.round(tf.reduce_sum(vals)) == 0


def test_euclidean_opposite():
    a = tf.convert_to_tensor([[0.0, 1.0], [0.0, -1.0]])
    d = EuclideanDistance()
    vals = d(a, a)
    assert tf.math.reduce_all(tf.shape(vals) == (2, 2))
    assert tf.round(tf.reduce_sum(vals)) == 4


def test_manhattan():
    a = tf.convert_to_tensor([
        [0.0, 0.0],
        [0.0, 1.0],
        [1.0, 1.0],
        [3.0, 0.0]
    ])
    d = ManhattanDistance()
    vals = d(a, a)
    expected = tf.constant([
        [0.0, 1.0, 2.0, 3.0],
        [1.0, 0.0, 1.0, 4.0],
        [2.0, 1.0, 0.0, 3.0],
        [3.0, 4.0, 3.0, 0.0]
    ])
    assert tf.math.reduce_all(tf.shape(vals) == (4, 4))
    assert tf.reduce_all(tf.math.equal(vals, expected))


def test_manhattan_key():
    a = tf.convert_to_tensor([
        [0.0, 0.0],
        [0.0, 1.0],
        [1.0, 1.0],
        [3.0, 0.0]
    ])
    b = tf.convert_to_tensor([
        [0.0, 1.0],
        [1.0, 1.0],
        [3.0, 0.0]
    ])
    d = ManhattanDistance()
    vals = d(a, b)
    expected = tf.constant([
        [1.0, 2.0, 3.0],
        [0.0, 1.0, 4.0],
        [1.0, 0.0, 3.0],
        [4.0, 3.0, 0.0]
    ])
    tf.assert_equal(vals, expected)

def test_manhattan_same():
    a = tf.convert_to_tensor([[1.0, 1.0], [1.0, 1.0]])
    d = ManhattanDistance()
    vals = d(a, a)
    assert tf.math.reduce_all(tf.shape(vals) == (2, 2))
    assert tf.round(tf.reduce_sum(vals)) == 0


def test_manhattan_opposite():
    a = tf.convert_to_tensor([[0.0, 1.0], [0.0, -1.0]])
    d = ManhattanDistance()
    vals = d(a, a)
    assert tf.math.reduce_all(tf.shape(vals) == (2, 2))
    assert tf.round(tf.reduce_sum(vals)) == 4


def test_snr_dist():
    """
    Comparing SNRDistance with simple loop based implementation
    of SNR distance.
    """
    num_inputs = 3
    dims = 5
    x = np.random.uniform(0, 1, (num_inputs, dims))

    # Computing SNR distance values using loop
    snr_pairs = []
    for i in range(num_inputs):
        row = []
        for j in range(num_inputs):
            dist = np.var(x[i] - x[j]) / np.var(x[i])
            row.append(dist)
        snr_pairs.append(row)
    snr_pairs = np.array(snr_pairs)

    x = tf.convert_to_tensor(x)
    snr_distances = SNRDistance()(x, x).numpy()
    assert np.all(snr_distances >= 0)
    diff = snr_distances - snr_pairs
    assert np.all(np.abs(diff) < 1e-4)


def test_snr_dist_key():
    """
    Comparing SNRDistance with simple loop based implementation
    of SNR distance for 2 different embedding tensors.
    """
    num_inputs = 3
    num_inputs2 = 2
    dims = 5
    x = np.random.uniform(0, 1, (num_inputs, dims))
    x2 = np.random.uniform(0, 1, (num_inputs2, dims))

    # Computing SNR distance values using loop
    snr_pairs = []
    for i in range(num_inputs):
        row = []
        for j in range(num_inputs2):
            dist = np.var(x[i] - x2[j]) / np.var(x[i])
            row.append(dist)
        snr_pairs.append(row)
    snr_pairs = np.array(snr_pairs)

    x = tf.convert_to_tensor(x)
    snr_distances = SNRDistance()(x, x2).numpy()
    assert np.all(snr_distances >= 0)
    diff = snr_distances - snr_pairs
    assert np.all(np.abs(diff) < 1e-4)



# Copyright 2021 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
from pathlib import Path
from typing import List, Optional, Sequence, Tuple

import numpy as np
import pandas as pd
import tensorflow as tf

from tensorflow_similarity.types import FloatTensor, PandasDataFrame, Tensor
from .store import Store


class MemoryStore(Store):
    """Efficient in-memory dataset store"""

    def __init__(self) -> None:
        # We are using a native python array in memory for its row speed.
        # Serialization / export relies on Arrow.
        self.labels: List[Optional[int]] = []
        self.embeddings: List[FloatTensor] = []
        self.data: List[Optional[Tensor]] = []
        self.num_items: int = 0
        pass

    def add(
        self,
        embedding: FloatTensor,
        label: Optional[int] = None,
        data: Optional[Tensor] = None,
    ) -> int:
        """Add an Embedding record to the key value store.

        Args:
            embedding: Embedding predicted by the model.

            label: Class numerical id. Defaults to None.

            data: Data associated with the embedding. Defaults to None.

        Returns:
            Associated record id.
        """
        idx = self.num_items
        self.labels.append(label)
        self.embeddings.append(embedding)
        self.data.append(data)
        self.num_items += 1
        return idx

    def batch_add(
        self,
        embeddings: Sequence[FloatTensor],
        labels: Optional[Sequence[int]] = None,
        data: Optional[Sequence[Tensor]] = None,
    ) -> List[int]:
        """Add a set of embedding records to the key value store.

        Args:
            embeddings: Embeddings predicted by the model.

            labels: Class numerical ids. Defaults to None.

            data: Data associated with the embeddings. Defaults to None.

        See:
            add() for what a record contains.

        Returns:
            List of associated record id.
        """
        idxs: List[int] = []
        for idx, embedding in enumerate(embeddings):
            label = None if labels is None else labels[idx]
            rec_data = None if data is None else data[idx]
            idxs.append(self.add(embedding, label, rec_data))
        return idxs

    def get(
        self, idx: int
    ) -> Tuple[FloatTensor, Optional[int], Optional[Tensor]]:
        """Get an embedding record from the key value store.

        Args:
            idx: Id of the record to fetch.

        Returns:
            record associated with the requested id.
        """

        return self.embeddings[idx], self.labels[idx], self.data[idx]

    def batch_get(
        self, idxs: Sequence[int]
    ) -> Tuple[List[FloatTensor], List[Optional[int]], List[Optional[Tensor]]]:
        """Get embedding records from the key value store.

        Args:
            idxs: ids of the records to fetch.

        Returns:
            List of records associated with the requested ids.
        """
        embeddings = []
        labels = []
        data = []
        for idx in idxs:
            e, l, d = self.get(idx)
            embeddings.append(e)
            labels.append(l)
            data.append(d)
        return embeddings, labels, data

    def size(self) -> int:
        "Number of record in the key value store."
        return self.num_items

    def save(self, path: str, compression: bool = True) -> None:
        """Serializes index on disk.

        Args:
            path: where to store the data.
            compression: Compress index data. Defaults to True.
        """
        # Writing to a buffer to avoid read error in np.savez when using GFile.
        # See: https://github.com/tensorflow/tensorflow/issues/32090
        io_buffer = io.BytesIO()
        if compression:
            np.savez_compressed(
                io_buffer,
                embeddings=self.embeddings,
                labels=np.array(self.labels),
                data=np.array(self.data),
            )
        else:
            np.savez(
                io_buffer,
                embeddings=self.embeddings,
                labels=np.array(self.labels),
                data=np.array(self.data),
            )

        with tf.io.gfile.GFile(self._make_fname(path), "wb+") as f:
            f.write(io_buffer.getvalue())

    def load(self, path: str) -> int:
        """load index on disk

        Args:
            path: which directory to use to store the index data.

        Returns:
           Number of records reloaded.
        """
        fname = self._make_fname(path, check_file_exit=True)
        with tf.io.gfile.GFile(fname, "rb") as gfp:
            data = np.load(gfp, allow_pickle=True)
        self.embeddings = list(data["embeddings"])
        self.labels = list(data["labels"])
        self.data = list(data["data"])
        self.num_items = len(self.embeddings)
        print("loaded %d records from %s" % (self.size(), path))
        return self.size()

    def _make_fname(self, path: str, check_file_exit: bool = False) -> str:
        p = Path(path)
        if not tf.io.gfile.exists(p):
            raise ValueError("Index path doesn't exist")
        fname = p / "index.npz"

        # only for loading
        if check_file_exit and not tf.io.gfile.exists(fname):
            raise ValueError("Index file not found")
        return str(fname)

    def to_data_frame(self, num_records: int = 0) -> PandasDataFrame:
        """Export data as a Pandas dataframe.

        Args:
            num_records: Number of records to export to the dataframe.
            Defaults to 0 (unlimited).

        Returns:
            pd.DataFrame: a pandas dataframe.
        """

        if not num_records:
            num_records = self.num_items

        data = {
            "embeddings": self.embeddings[:num_records],
            "data": self.data[:num_records],
            "lables": self.labels[:num_records],
        }

        # forcing type from Any to PandasFrame
        df: PandasDataFrame = pd.DataFrame.from_dict(data)
        return df



import math

import pytest
import tensorflow as tf

from tensorflow_similarity import utils
from tensorflow_similarity import types


@pytest.fixture
def lookups():
    lookups = []
    for i in range(2):
        lookup_set = []
        for j in range(2):
            label = j + i * 2
            lookup_set.append(
                types.Lookup(rank=j, distance=float(j), label=label))

        lookups.append(lookup_set)

    return lookups


def test_is_tensor():
    assert utils.is_tensor_or_variable(tf.constant([0]))


def test_is_variable():
    assert utils.is_tensor_or_variable(tf.Variable(1.0))


def test_is_not_tensor_or_variable():
    assert not utils.is_tensor_or_variable([0])


def test_unpack_lookup_labels(capsys, lookups):
    unpacked = utils.unpack_lookup_labels(lookups, dtype="int32")
    expected = tf.constant([[0, 1], [2, 3]], dtype="int32")

    assert tf.reduce_all(tf.math.equal(unpacked, expected))

    captured = capsys.readouterr()
    assert captured.out == ""


def test_unpack_lookup_labels_uneven_lookup_sets(capsys, lookups):
    # Add an extra label to the second lookup set
    lookups[1].append(types.Lookup(rank=3, distance=math.inf, label=4))

    unpacked = utils.unpack_lookup_labels(lookups, dtype="int32")
    expected = tf.constant([[0, 1, 0x7FFFFFFF], [2, 3, 4]], dtype="int32")

    assert tf.reduce_all(tf.math.equal(unpacked, expected))

    captured = capsys.readouterr()
    msg = ("WARNING: 1 lookup sets are shorter than the max lookup set "
           "length. Imputing 0x7FFFFFFF for the missing label lookups.\n")

    assert captured.out == msg


def test_unpack_lookup_distances(capsys, lookups):
    unpacked = utils.unpack_lookup_distances(lookups, dtype="float32")
    expected = tf.constant([[0.0, 1.0], [0.0, 1.0]], dtype="float32")

    assert tf.reduce_all(tf.math.equal(unpacked, expected))

    captured = capsys.readouterr()
    assert captured.out == ""


def test_unpack_lookup_distances_uneven_lookup_sets(capsys, lookups):
    # Add an extra label to the second lookup set
    lookups[1].append(types.Lookup(rank=3, distance=2.0, label=4))

    unpacked = utils.unpack_lookup_distances(lookups, dtype="float32")
    expected = tf.constant([[0.0, 1.0, math.inf], [0.0, 1.0, 2.0]],
                           dtype="float32")

    assert tf.reduce_all(tf.math.equal(unpacked, expected))

    captured = capsys.readouterr()
    msg = ("WARNING: 1 lookup sets are shorter than the max lookup set "
           "length. Imputing math.inf for the missing distance lookups.\n")

    assert captured.out == msg


def test_same_length_rows_check_same_length():
    x = tf.ragged.constant([[0, 1], [0, 2], [0, 3]])
    assert utils._same_length_rows(x)


def test_same_length_rows_check_different_lengths():
    x = tf.ragged.constant([[0], [0, 2], [0, 2, 3]])
    assert not utils._same_length_rows(x)


def test_count_of_small_lookup_sets_ragged():
    x = tf.ragged.constant([[0], [0, 2], [0, 2, 3]])
    counts = utils._count_of_small_lookup_sets(x)
    assert counts == 2


def test_count_of_small_lookup_sets_all_same_length():
    x = tf.ragged.constant([[0, 2], [0, 2], [0, 2]])
    counts = utils._count_of_small_lookup_sets(x)
    assert counts == 0



# Copyright 2021 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import abc
from tensorflow_similarity.types import Tensor

# from tensorflow import Tensor
from typing import List


class Augmenter(abc.ABC):
    @abc.abstractmethod
    def augment(
        self,
        x: Tensor,
        y: Tensor,
        num_augmentations_per_example: int,
        is_warmup: bool,
    ) -> List[Tensor]:
        pass

    def __call__(
        self,
        x: Tensor,
        y: Tensor,
        num_augmentations_per_example: int,
        is_warmup: bool,
    ) -> List[Tensor]:
        return self.augment(x, y, num_augmentations_per_example, is_warmup)



# coding=utf-8
# Copyright 2020 The SimCLR Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific simclr governing permissions and
# limitations under the License.
# ==============================================================================
"""Data preprocessing and augmentation."""

import os
from typing import List, Optional

import tensorflow as tf

from tensorflow_similarity.augmenters import Augmenter
from tensorflow_similarity.types import Tensor

from .augmentation_utils.color_jitter import random_color_jitter
from .augmentation_utils.cropping import center_crop, random_crop_with_resize


def simclr_training_augmentation(
    image: Tensor,
    height: int,
    width: int,
    color_distort: bool = True,
    jitter_stength: float = 1.0,
    crop: bool = True,
    flip: bool = True,
    impl: str = "multiplicative",
) -> Tensor:
    """SimCLR Preprocesses the given image for training.

    Args:
      image: `Tensor` representing an image of arbitrary size.
      height: Height of output image.
      width: Width of output image.
      color_distort: Whether to apply the color distortion.
      crop: Whether to crop the image.
      flip: Whether or not to flip left and right of an image.
      impl: 'additive' or 'multiplicative'.  Whether to use simclrv1 or simclrv2's
          version of random brightness.

    Returns:
      A preprocessed image `Tensor`.
    """
    if crop:
        image = random_crop_with_resize(image, height, width)
    if flip:
        image = tf.image.random_flip_left_right(image)
    if color_distort:
        image = random_color_jitter(image, strength=jitter_stength, impl=impl)
    image = tf.reshape(image, [height, width, 3])
    image = tf.clip_by_value(image, 0.0, 1.0)
    return image


def simclr_eval_augmentation(
    image: Tensor,
    height: int,
    width: int,
    crop: bool = True,
    crop_proportion: float = 0.875,
) -> Tensor:
    """Preprocesses the given image for evaluation.

    Args:
      image: `Tensor` representing an image of arbitrary size.
      height: Height of output image.
      width: Width of output image.
      crop: Whether or not to (center) crop the test images.

    Returns:
      A preprocessed image `Tensor`.
    """
    if crop:
        image = center_crop(
            image, height, width, crop_proportion=crop_proportion
        )
    image = tf.reshape(image, [height, width, 3])
    image = tf.clip_by_value(image, 0.0, 1.0)
    return image


class SimCLRAugmenter(Augmenter):
    """SimCLR augmentation pipeline as defined in
    [Big Self-Supervised Models are Strong Semi-Supervised Learners](https://arxiv.org/abs/2006.10029)

    code adapted from [orignal github](https://github.com/google-research/simclr/tree/master/tf2)
    """

    def __init__(
        self,
        height: int,
        width: int,
        is_training: bool = True,
        color_distort: bool = True,
        jitter_stength: float = 1.0,
        crop: bool = True,
        eval_crop_proportion: float = 0.875,  # imagenet standard
        flip: bool = True,
        version: str = "v2",
        num_cpu: Optional[int] = os.cpu_count(),
    ):

        self.width = width
        self.height = height
        self.is_training = is_training
        self.color_distort = color_distort
        self.jitter_stength = jitter_stength
        self.crop = crop
        self.eval_crop_proportion = eval_crop_proportion
        self.flip = flip
        self.version = version
        self.num_cpu = num_cpu

        if version == "v2":
            self.impl = "multiplicative"
        else:
            self.impl = "additive"

        if self.is_training:
            self.augment_img = self._train_augment_img
        else:
            self.augment_img = self._eval_augment_img

    @tf.function
    def augment(
        self, x: Tensor, y: Tensor, num_views: int, is_warmup: bool
    ) -> List[Tensor]:

        with tf.device("/cpu:0"):
            inputs = tf.stack(x)
            inputs = tf.cast(inputs, dtype="float32") / 255.0
            views = []

            for _ in range(num_views):
                # multi-cor augementations
                view = tf.map_fn(
                    self.augment_img, inputs, parallel_iterations=self.num_cpu
                )
                views.append(view)
        return views

    def _train_augment_img(self, img: Tensor) -> Tensor:
        return simclr_training_augmentation(
            img,
            self.height,
            self.width,
            self.color_distort,
            self.jitter_stength,
            self.crop,
            self.flip,
            self.impl,
        )

    def _eval_augment_img(self, img: Tensor) -> Tensor:
        return simclr_eval_augmentation(
            img, self.height, self.width, self.crop, self.eval_crop_proportion
        )



# Copyright 2021 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, MutableMapping, Sequence, Union

import numpy as np
import tensorflow as tf
from tqdm.auto import tqdm

from tensorflow_similarity.classification_metrics import ClassificationMetric
from tensorflow_similarity.matchers import (ClassificationMatch,
                                            make_classification_matcher)
from tensorflow_similarity.retrieval_metrics import RetrievalMetric
from tensorflow_similarity.retrieval_metrics.utils import compute_match_mask
from tensorflow_similarity.types import (CalibrationResults, FloatTensor,
                                         IntTensor, Lookup)
from tensorflow_similarity.utils import (unpack_lookup_distances,
                                         unpack_lookup_labels)

from .evaluator import Evaluator


class MemoryEvaluator(Evaluator):
    """In memory index performance evaluation and classification."""

    def evaluate_retrieval(
        self,
        target_labels: Sequence[int],
        lookups: Sequence[Sequence[Lookup]],
        retrieval_metrics: Sequence[RetrievalMetric],
        distance_rounding: int = 8,
    ) -> Dict[str, np.ndarray]:
        """Evaluates lookup performances against a supplied set of metrics

        Args:
            target_labels: Sequence of the expected labels to match.

            lookups: Sequence of lookup results as produced by the
            `Index().batch_lookup()` method.

            retrieval_metrics: Sequence of `RetrievalMetric()` to evaluate
            lookup matches against.

            distance_rounding: How many digit to consider to decide if
            the distance changed. Defaults to 8.

        Returns:
            Dictionary of metric results where keys are the metric names and
            values are the metrics values.
        """
        # we also convert to np.ndarray first to avoid a slow down if
        # convert_to_tensor is called on a List.
        query_labels = tf.convert_to_tensor(np.array(target_labels))

        # data preparation: flatten and rounding
        # lookups will be shape(num_queries, num_neighbors)
        # distances will be len(num_queries x num_neighbors)
        nn_labels = unpack_lookup_labels(lookups, dtype=query_labels.dtype)
        # TODO(ovallis): The float type should be derived from the model.
        distances = unpack_lookup_distances(
            lookups, dtype="float32", distance_rounding=distance_rounding
        )

        lookup_set_size = tf.shape(nn_labels)[1]
        for m in retrieval_metrics:
            if lookup_set_size < m.k:
                raise ValueError(
                    f"Each query example returned {lookup_set_size} "
                    f"neighbors, but retrieval metric {m.name} "
                    f"requires the K >= {m.k}."
                )

        match_mask = compute_match_mask(query_labels, nn_labels)

        # compute metrics
        evaluation = {}
        for m in retrieval_metrics:
            res = m.compute(
                query_labels=query_labels,
                lookup_labels=nn_labels,
                lookup_distances=distances,
                match_mask=match_mask,
            )
            evaluation[m.name] = res.numpy()

        return evaluation

    def evaluate_classification(
        self,
        query_labels: IntTensor,
        lookup_labels: IntTensor,
        lookup_distances: FloatTensor,
        distance_thresholds: FloatTensor,
        metrics: Sequence[ClassificationMetric],
        matcher: Union[str, ClassificationMatch],
        distance_rounding: int = 8,
        verbose: int = 1,
    ) -> Dict[str, np.ndarray]:
        """Evaluate the classification performance.

        Compute the classification metrics given a set of queries, lookups, and
        distance thresholds.

        Args:
            query_labels: Sequence of expected labels for the lookups.

            lookup_labels: A 2D tensor where the jth row is the labels
            associated with the set of k neighbors for the jth query.

            lookup_distances: A 2D tensor where the jth row is the distances
            between the jth query and the set of k neighbors.

            distance_thresholds: A 1D tensor denoting the distances points at
            which we compute the metrics.

            metrics: The set of classification metrics.

            matcher: {'match_nearest', 'match_majority_vote'} or
            ClassificationMatch object. Defines the classification matching,
            e.g., match_nearest will count a True Positive if the query_label
            is equal to the label of the nearest neighbor and the distance is
            less than or equal to the distance threshold.

            distance_rounding: How many digit to consider to
            decide if the distance changed. Defaults to 8.

            verbose: Be verbose. Defaults to 1.

        Returns:
            A Mapping from metric name to the list of values computed for each
            distance threshold.
        """
        matcher = make_classification_matcher(matcher)
        matcher.compile(distance_thresholds=distance_thresholds)

        # compute the tp, fp, tn, fn counts
        matcher.compute_count(
            query_labels=query_labels,
            lookup_labels=lookup_labels,
            lookup_distances=lookup_distances,
        )

        # evaluating performance as distance value increase
        if verbose:
            pb = tqdm(total=len(metrics), desc="Evaluating")

        # evaluating performance as distance value increase
        results: Dict[str, np.ndarray] = {
            "distance": distance_thresholds.numpy()
        }
        for m in metrics:
            res = m.compute(
                tp=matcher.tp,
                fp=matcher.fp,
                tn=matcher.tn,
                fn=matcher.fn,
                count=matcher.count,
            )
            results[m.name] = res.numpy()

            if verbose:
                pb.update()

        if verbose:
            pb.close()

        return results

    def calibrate(
        self,
        target_labels: Sequence[int],
        lookups: Sequence[Sequence[Lookup]],
        thresholds_targets: MutableMapping[str, float],
        calibration_metric: ClassificationMetric,
        matcher: Union[str, ClassificationMatch],
        extra_metrics: Sequence[ClassificationMetric] = [],
        distance_rounding: int = 8,
        metric_rounding: int = 6,
        verbose: int = 1,
    ) -> CalibrationResults:
        """Computes the distances thresholds that the classification must match to
        meet a fixed target.

        Args:
            target_labels: Sequence of expected labels for the lookups.

            lookup: Sequence of lookup results as produced by the
            `Index.batch_lookup()` method.

            thresholds_targets: classification metrics thresholds that are
            targeted. The function will find the closed distance value.

            calibration_metric: Classification metric used for calibration.

            matcher: {'match_nearest', 'match_majority_vote'} or
            ClassificationMatch object. Defines the classification matching,
            e.g., match_nearest will count a True Positive if the query_label
            is equal to the label of the nearest neighbor and the distance is
            less than or equal to the distance threshold.

            extra_metrics: Additional classification metrics that should be
            computed and reported as part of the calibration. Defaults to [].

            distance_rounding: How many digit to consider to
            decide if the distance changed. Defaults to 8.

            metric_rounding: How many digit to consider to decide if
            the metric changed. Defaults to 6.

            verbose: Be verbose. Defaults to 1.
        Returns:
            CalibrationResults containing the thresholds and cutpoints Dicts.

        Raises:
            ValueError: lookupds must not be empty.
        """
        # TODO (ovallis): Assert if index is empty, or if the lookup is empty.
        if len(lookups) == 0:
            raise ValueError(
                "lookups must not be empty. Is there no data in the index?"
            )

        # making a single list of metrics
        # Need expl covariance problem
        combined_metrics = list(extra_metrics)
        combined_metrics.append(calibration_metric)

        # we also convert to np.ndarray first to avoid a slow down if
        # convert_to_tensor is called on a List.
        query_labels = tf.convert_to_tensor(np.array(target_labels))

        # data preparation: flatten and rounding
        # lookups will be shape(num_queries, num_neighbors)
        # distances will be len(num_queries x num_neighbors)
        # TODO(ovallis): The float type should be derived from the model.
        lookup_distances = unpack_lookup_distances(
            lookups, dtype="float32", distance_rounding=distance_rounding
        )
        lookup_labels = unpack_lookup_labels(lookups, dtype=query_labels.dtype)

        # the unique set of distance values sorted ascending
        unique_distances, _ = tf.unique(tf.reshape(lookup_distances, (-1)))
        distance_thresholds = tf.sort(unique_distances)

        results = self.evaluate_classification(
            query_labels=query_labels,
            lookup_labels=lookup_labels,
            lookup_distances=lookup_distances,
            distance_thresholds=distance_thresholds,
            metrics=combined_metrics,
            matcher=matcher,
            distance_rounding=distance_rounding,
            verbose=verbose,
        )

        cutpoints: Dict[str, Dict[str, Union[str, float]]] = {}

        cutpoints["optimal"] = self._optimal_cutpoint(
            results, calibration_metric
        )

        for name, value in thresholds_targets.items():
            target_cp = self._target_cutpoints(
                results, calibration_metric, name, value
            )
            if target_cp:
                cutpoints[name] = target_cp

        # Add the calibration metric as 'value' in the thresholds dict.
        results['value'] = results[calibration_metric.name]

        return CalibrationResults(cutpoints=cutpoints, thresholds=results)

    def _optimal_cutpoint(
        self,
        metrics: Dict[str, np.ndarray],
        calibration_metric: ClassificationMetric,
    ) -> Dict[str, Union[str, float]]:
        """Compute the optimal distance threshold for the calibration metric.

        Args:
            metrics: A mapping from metric name to a list of metric values
              computed for each unique distance observed in the calibration
              dataset. This dict should also include a distance key mapping
              to the list of observed distances. We expect the distances to be
              ascending, and for all lists to be the same length where each
              index in the list corresponds to a distance in the distance list.
            calibration_metric: The ClassificationMetric used for calibration.

        Returns:
            A Dict of the metric values at the calibrated distance. This also
              includes the cutpoint name, the distance, and the value of the
              calibration metric.

              ```
              {
                  'name': 'optimal', # Cutpoint name
                  'value': 0.99,     # Calibration metric at the cutpoint
                  'distance': 0.1,   # Calibrated distance
                  'precision': 0.99, # Here, we calibrated using precision
                  'f1': 0.4,         # We also computed F1 at this point
              }
              ```
        """
        if calibration_metric.maximize:
            idx = self._last_argmax(metrics[calibration_metric.name])
        else:
            idx = self._last_argmin(metrics[calibration_metric.name])

        optimal_cp = {
            "name": "optimal",
            "value": metrics[calibration_metric.name][idx].item(),
        }
        for metric_name in metrics.keys():

            optimal_cp[metric_name] = metrics[metric_name][idx].item()

        return optimal_cp

    def _target_cutpoints(
        self,
        metrics: Dict[str, np.ndarray],
        calibration_metric: ClassificationMetric,
        target_name: str,
        target_value: float,
    ) -> Dict[str, Union[str, float]]:
        """Compute the distance at the target metric for the calibration metric.

        Args:
            metrics: A mapping from metric name to a list of metric values
              computed for each unique distance observed in the calibration
              dataset. This dict should also include a distance key mapping
              to the list of observed distances. We expect the distances to be
              ascending, and for all lists to be the same length where each
              index in the list corresponds to a distance in the distance list.
            calibration_metric: The ClassificationMetric used for calibration.
            target_name: The name for the target cutpoint.
            target_value: The target metric value.

        Returns:
            A Dict of the metric values at the calibrated distance. This also
              includes the cutpoint name, the distance, and the value of the
              calibration metric.

              ```
              {
                  'name': '0.90',     # Target cutpoint name.
                  'value': 0.901,     # Closest metric value at or above the
                                      # target cutpoint, assuming we are
                                      # maximizing the calibration metric.
                  'distance': 0.1,    # Calibrated distance.
                  'precision': 0.901, # Here, we calibrated using precision.
                  'f1': 0.4,          # We also computed F1 at this point.
              }
              ```
        """
        indicators = np.where(metrics[calibration_metric.name] >= target_value)[
            0
        ]
        target_cp: Dict[str, Union[str, float]] = {}

        if indicators.size > 0:
            if calibration_metric.increasing:
                # Take the first index above the target if the metric is increasing
                idx = indicators[0]
            else:
                # Take the last index above the target if the metric is decreasing
                idx = indicators[-1]

            target_cp["name"] = target_name
            target_cp["value"] = metrics[calibration_metric.name][idx].item()
            for metric_name in metrics.keys():
                target_cp[metric_name] = metrics[metric_name][idx].item()

        return target_cp

    def _last_argmax(self, x: np.ndarray) -> int:
        """The index of the last occurrence of the max value.

        In case of multiple occurrences of the maximum values, the index
        corresponding to the last occurrence is returned.

        Args:
            A 1D np.ndarray or List[float].

        Returns:
            The index of the last occurrence of the max value.
        """
        revx = x[::-1]
        return (len(x) - np.argmax(revx) - 1).item()

    def _last_argmin(self, x: np.ndarray) -> int:
        """The index of the last occurrence of the min value.

        In case of multiple occurrences of the minimum values, the index
        corresponding to the last occurrence is returned.

        Args:
            A 1D np.ndarray or List[float].

        Returns:
            The index of the last occurrence of the min value.
        """
        revx = x[::-1]
        return (len(x) - np.argmin(revx) - 1).item()

    def _is_lower(self, curr, prev, equal=False):
        if equal:
            return curr <= prev
        return curr < prev

    def _is_higher(self, curr, prev, equal=False):
        if equal:
            return curr >= prev
        return curr > prev



# Copyright 2021 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Specialized Similarity `keras.layers`"""
import math
from typing import Any, Dict, Optional

import tensorflow as tf
from tensorflow.keras import layers
from tensorflow.python.keras.utils import conv_utils

from .types import FloatTensor, IntTensor


@tf.keras.utils.register_keras_serializable(package="Similarity")
class MetricEmbedding(layers.Dense):
    """L2 Normalized `Dense` layer.

    This layer is usually used as output layer, especially when using cosine
    distance as the similarity metric.
    """
    def call(self, inputs: FloatTensor) -> FloatTensor:
        x = super().call(inputs)
        normed_x: FloatTensor = tf.math.l2_normalize(x, axis=1)
        return normed_x


class GeneralizedMeanPooling(layers.Layer):
    def __init__(
        self,
        p: float = 3.0,
        data_format: Optional[str] = None,
        keepdims: bool = False,
        **kwargs
    ) -> None:
        super().__init__(**kwargs)

        self.p = p
        self.data_format = conv_utils.normalize_data_format(data_format)
        self.keepdims = keepdims

        if tf.math.abs(self.p) < 0.00001:
            self.compute_mean = self._geometric_mean
        elif self.p == math.inf:
            self.compute_mean = self._pos_inf
        elif self.p == -math.inf:
            self.compute_mean = self._neg_inf
        else:
            self.compute_mean = self._generalized_mean

    def compute_output_shape(self, input_shape: IntTensor) -> IntTensor:
        output_shape: IntTensor = self.gap.compute_output_shape(input_shape)
        return output_shape

    def call(self, inputs: FloatTensor) -> FloatTensor:
        raise NotImplementedError

    def _geometric_mean(self, x):
        x = tf.math.log(x)
        x = self.gap(x)
        return tf.math.exp(x)

    def _generalized_mean(self, x):
        x = tf.math.pow(x, self.p)
        x = self.gap(x)
        return tf.math.pow(x, 1.0 / self.p)

    def _pos_inf(self, x):
        raise NotImplementedError

    def _neg_inf(self, x):
        return self._pos_inf(x * -1) * -1

    def get_config(self) -> Dict[str, Any]:
        config = {
            "p": self.p,
            "data_format": self.data_format,
            "keepdims": self.keepdims,
        }
        base_config = super().get_config()
        return {**base_config, **config}


@tf.keras.utils.register_keras_serializable(package="Similarity")
class GeneralizedMeanPooling1D(GeneralizedMeanPooling):
    r"""Computes the Generalized Mean of each channel in a tensor.

    $$
    \textbf{e} = \left[\left(\frac{1}{|\Omega|}\sum_{u\in{\Omega}}x^{p}_{cu}\right)^{\frac{1}{p}}\right]_{c=1,\cdots,C}
    $$

    The Generalized Mean (GeM) provides a parameter `p` that sets an exponent
    enabling the pooling to increase or decrease the contrast between salient
    features in the feature map.

    The pooling is equal to GlobalAveragePooling1D when `p` is 1.0 and equal
    to MaxPool1D when `p` is `inf`.

    This implementation shifts the feature map values such that the minimum
    value is equal to 1.0, then computes the mean pooling, and finally shifts
    the values back. This ensures that all values are positive as the
    generalized mean is only valid over positive real values.

    Args:
      p: Set the power of the mean. A value of 1.0 is equivalent to the
        arithmetic mean, while a value of `inf` is equivalent to MaxPool2D.
        Note, math.inf, -math.inf, and 0.0 are all supported, as well as most
        positive and negative values of `p`. However, large positive values for
        `p` may lead to overflow. In practice, math.inf should be used for any
        `p` larger than > 25.
      data_format: One of `channels_last` (default) or `channels_first`. The
        ordering of the dimensions in the inputs.  `channels_last`
        corresponds to inputs with shape `(batch, steps, features)` while
        `channels_first` corresponds to inputs with shape
        `(batch, features, steps)`.
      keepdims: A boolean, whether to keep the temporal dimension or not.
        If `keepdims` is `False` (default), the rank of the tensor is reduced
        for spatial dimensions.  If `keepdims` is `True`, the temporal
        dimension are retained with length 1.  The behavior is the same as
        for `tf.reduce_max` or `np.max`.

    Input shape:
      - If `data_format='channels_last'`:
        3D tensor with shape:
        `(batch_size, steps, features)`
      - If `data_format='channels_first'`:
        3D tensor with shape:
        `(batch_size, features, steps)`
    Output shape:
      - If `keepdims`=False:
        2D tensor with shape `(batch_size, features)`.
      - If `keepdims`=True:
        - If `data_format='channels_last'`:
          3D tensor with shape `(batch_size, 1, features)`
        - If `data_format='channels_first'`:
          3D tensor with shape `(batch_size, features, 1)`
    """

    def __init__(
        self,
        p: float = 3.0,
        data_format: Optional[str] = None,
        keepdims: bool = False,
        **kwargs
    ) -> None:
        super().__init__(
            p=p, data_format=data_format, keepdims=keepdims, **kwargs
        )

        self.input_spec = layers.InputSpec(ndim=3)
        self.gap = layers.GlobalAveragePooling1D(
            data_format=data_format, keepdims=keepdims
        )
        self.step_axis = 1 if self.data_format == "channels_last" else 2

    def call(self, inputs: FloatTensor) -> FloatTensor:
        x = inputs
        if self.data_format == "channels_last":
            mins = tf.math.reduce_min(x, axis=self.step_axis)
            x_offset = x - mins[:, tf.newaxis, :] + 1
            if self.keepdims:
                mins = mins[:, tf.newaxis, :]
        else:
            mins = tf.math.reduce_min(x, axis=self.step_axis)
            x_offset = x - mins[:, :, tf.newaxis] + 1
            if self.keepdims:
                mins = mins[:, :, tf.newaxis]

        x_offset = self.compute_mean(x_offset)
        x = x_offset + mins - 1

        return x

    def _pos_inf(self, x):
        mpl = layers.GlobalMaxPool1D(data_format=self.data_format)
        x = mpl(x)
        return x


@tf.keras.utils.register_keras_serializable(package="Similarity")
class GeneralizedMeanPooling2D(GeneralizedMeanPooling):
    r"""Computes the Generalized Mean of each channel in a tensor.

    $$
    \textbf{e} = \left[\left(\frac{1}{|\Omega|}\sum_{u\in{\Omega}}x^{p}_{cu}\right)^{\frac{1}{p}}\right]_{c=1,\cdots,C}
    $$

    The Generalized Mean (GeM) provides a parameter `p` that sets an exponent
    enabling the pooling to increase or decrease the contrast between salient
    features in the feature map.

    The pooling is equal to GlobalAveragePooling2D when `p` is 1.0 and equal
    to MaxPool2D when `p` is `inf`.

    This implementation shifts the feature map values such that the minimum
    value is equal to 1.0, then computes the mean pooling, and finally shifts
    the values back. This ensures that all values are positive as the
    generalized mean is only valid over positive real values.

    Args:
      p: Set the power of the mean. A value of 1.0 is equivalent to the
        arithmetic mean, while a value of `inf` is equivalent to MaxPool2D.
        Note, math.inf, -math.inf, and 0.0 are all supported, as well as most
        positive and negative values of `p`. However, large positive values for
        `p` may lead to overflow. In practice, math.inf should be used for any
        `p` larger than > 25.
      data_format: One of `channels_last` (default) or `channels_first`. The
        ordering of the dimensions in the inputs.  `channels_last`
        corresponds to inputs with shape `(batch, steps, features)` while
        `channels_first` corresponds to inputs with shape
        `(batch, features, steps)`.
      keepdims: A boolean, whether to keep the temporal dimension or not.
        If `keepdims` is `False` (default), the rank of the tensor is reduced
        for spatial dimensions.  If `keepdims` is `True`, the temporal
        dimension are retained with length 1.  The behavior is the same as
        for `tf.reduce_max` or `np.max`.

    Input shape:
      - If `data_format='channels_last'`:
        3D tensor with shape:
        `(batch_size, steps, features)`
      - If `data_format='channels_first'`:
        3D tensor with shape:
        `(batch_size, features, steps)`
    Output shape:
      - If `keepdims`=False:
        2D tensor with shape `(batch_size, features)`.
      - If `keepdims`=True:
        - If `data_format='channels_last'`:
          3D tensor with shape `(batch_size, 1, features)`
        - If `data_format='channels_first'`:
          3D tensor with shape `(batch_size, features, 1)`
    """

    def __init__(
        self,
        p: float = 3.0,
        data_format: Optional[str] = None,
        keepdims: bool = False,
        **kwargs
    ) -> None:
        super().__init__(
            p=p, data_format=data_format, keepdims=keepdims, **kwargs
        )

        self.input_spec = layers.InputSpec(ndim=4)
        self.gap = layers.GlobalAveragePooling2D(data_format, keepdims)

    def call(self, inputs: FloatTensor) -> FloatTensor:
        x = inputs
        if self.data_format == "channels_last":
            mins = tf.math.reduce_min(x, axis=[1, 2])
            x_offset = x - mins[:, tf.newaxis, tf.newaxis, :] + 1
            if self.keepdims:
                mins = mins[:, tf.newaxis, tf.newaxis, :]
        else:
            mins = tf.math.reduce_min(x, axis=[2, 3])
            x_offset = x - mins[:, :, tf.newaxis, tf.newaxis] + 1
            if self.keepdims:
                mins = mins[:, :, tf.newaxis, tf.newaxis]

        x_offset = self.compute_mean(x_offset)
        x = x_offset + mins - 1

        return x

    def _pos_inf(self, x):
        if self.data_format == "channels_last":
            pool_size = (x.shape[1], x.shape[2])
        else:
            pool_size = (x.shape[2], x.shape[3])
        mpl = layers.MaxPool2D(
            pool_size=pool_size, data_format=self.data_format
        )
        x = mpl(x)
        if not self.keepdims:
            if self.data_format == "channels_last":
                x = tf.reshape(x, (x.shape[0], x.shape[3]))
            else:
                x = tf.reshape(x, (x.shape[0], x.shape[1]))
        return x


class ActivationStdLoggingLayer(layers.Layer):
    """Computes the mean std of the activations of a layer.

    x = reduce_std(l2_normalize(inputs, axis=0), axis=-1)

    And then aggregate the per-batch mean of x over each epoch.
    """

    def __init__(self, name, **kwargs):
        super().__init__(name=name, **kwargs)

    def call(self, inputs):
        x = tf.math.l2_normalize(inputs, axis=-1)
        x = tf.math.reduce_std(x, axis=0)
        self.add_metric(x, name=self.name, aggregation="mean")
        return inputs  # Pass-through layer.



# Copyright 2021 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""VicReg Loss
    Cross-batch memory for embedding learning.
    https://arxiv.org/abs/1912.06798
"""
from typing import Any, Callable, Dict, Optional

import tensorflow as tf

from tensorflow_similarity.losses import MetricLoss
from tensorflow_similarity.types import FloatTensor


def _add_memory_variable(tensor):
    """Creates an empty variable with same shape and dtype as `tensor`."""
    shape_no_batch = tensor.shape[1:]
    dtype = tensor.dtype
    init_value = tf.constant([], shape=[0, *shape_no_batch], dtype=dtype)
    var = tf.Variable(
        initial_value=init_value,
        shape=[None, *shape_no_batch],
        dtype=dtype,
        trainable=False,
    )
    return var


@tf.keras.utils.register_keras_serializable(package="Similarity")
class XBM(MetricLoss):
    """Cross-batch memory wrapper for MetricLoss instances.

    Maintains a memory queue of past embedding batches. Batch embeddings are
    paired with all embeddings in the memory queue, and loss is calculated from
    these pairs.

    Reference

    Wang, Xun, et al. "Cross-batch memory for embedding learning."
    https://arxiv.org/abs/1912.06798

    Standalone usage:

    >>> loss = tensorflow_similarity.losses.MultiSimilarityLoss()
    >>> loss = XBM(loss, memory_size=1000, warmup_steps=100)
    >>> y_pred = tf.random.uniform(shape=[4, 16])
    >>> y_true = tf.constant([[1], [2], [2], [3]])
    >>> loss(y_true, y_pred)
    <tf.Tensor: shape=(), dtype=float32, numpy=0.3740389>

    Args:
      loss: MetricLoss instance to use for computing loss.
      memory_size: Integer specifying the number of past embeddings to
        maintain in the memory queue.
      warmup_steps: Integer specifying the number of warmup steps where
        loss is calculated without using the memory queue.

    Returns:
      Loss value for cross-batched pairs

    NOTE:
    This will trigger multiple tf.function retracings if called multiple
    times in Eager mode, because of the dynamic size of the memory queue.

    """

    def __init__(
        self,
        loss: MetricLoss,
        memory_size: int,
        warmup_steps: int = 0,
        reduction: Callable = tf.keras.losses.Reduction.AUTO,
        name: Optional[str] = None,
        **kwargs,
    ):
        super().__init__(loss.fn, reduction, name, **kwargs)

        self.loss = loss
        self.distance = loss.distance
        self._fn_kwargs.update(loss._fn_kwargs)

        # Diagonal is removed from positive pair mask because the batch is
        # always contained in the beginning of the memory.
        self._fn_kwargs["remove_diagonal"] = True

        self.memory_size = memory_size
        self.warmup_steps = warmup_steps
        self._y_true_memory = None
        self._y_pred_memory = None
        self._total_steps = tf.Variable(0, dtype=tf.int64, trainable=False)

    def call(self, y_true: FloatTensor, y_pred: FloatTensor) -> FloatTensor:
        with tf.device("/cpu:0"):
            # Build memory from first batch
            if self._y_true_memory is None:
                self._y_true_memory = _add_memory_variable(y_true)
            if self._y_pred_memory is None:
                self._y_pred_memory = _add_memory_variable(y_pred)

        # Enqueue (concat batch to beginning of memory)
        y_true_mem = tf.concat([y_true, self._y_true_memory], axis=0)
        y_pred_mem = tf.concat([y_pred, self._y_pred_memory], axis=0)

        # Dequeue (truncate to memory size limit)
        y_true_mem = y_true_mem[: self.memory_size]
        y_pred_mem = y_pred_mem[: self.memory_size]

        def _xbm_step():
            # Update memory with new values
            self._y_true_memory.assign(y_true_mem)
            self._y_pred_memory.assign(y_pred_mem)
            return y_true_mem, y_pred_mem

        def _warmup_step():
            # Memory is not updated during warmup steps, so it remains empty.
            # Therefore this corresponds to the original batch
            return y_true_mem, y_pred_mem

        self._total_steps.assign_add(1)
        y_true_mem, y_pred_mem = tf.cond(
            self._total_steps > self.warmup_steps,
            true_fn=_xbm_step,
            false_fn=_warmup_step,
        )

        loss: FloatTensor = self.fn(
            y_true, y_pred, y_true_mem, y_pred_mem, **self._fn_kwargs
        )
        return loss

    def get_config(self) -> Dict[str, Any]:
        config = {
            "loss": tf.keras.utils.serialize_keras_object(self.loss),
            "memory_size": self.memory_size,
            "warmup_steps": self.warmup_steps,
        }

        base_config = super().get_config()
        return dict(list(base_config.items()) + list(config.items()))

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> MetricLoss:
        config["loss"] = tf.keras.utils.deserialize_keras_object(config["loss"])
        return cls(**config)



import tensorflow as tf
import numpy as np
from tensorflow_similarity.losses import TripletLoss, MultiSimilarityLoss
from tensorflow_similarity.losses import PNLoss
from tensorflow_similarity.losses import SoftNearestNeighborLoss


# [triplet loss]
from tensorflow_similarity.losses.xbm_loss import XBM


def test_triplet_loss_serialization():
    loss = TripletLoss()
    config = loss.get_config()
    print(config)
    loss2 = TripletLoss.from_config(config)
    assert loss.name == loss2.name
    assert loss.distance == loss2.distance


def triplet_hard_loss_np(labels, embedding, margin, dist_func, soft=False):

    num_data = embedding.shape[0]
    # Reshape labels to compute adjacency matrix.
    labels_reshaped = np.reshape(labels.astype(np.float32),
                                 (labels.shape[0], 1))

    adjacency = np.equal(labels_reshaped, labels_reshaped.T)
    pdist_matrix = dist_func(embedding)
    loss_np = 0.0
    for i in range(num_data):
        pos_distances = []
        neg_distances = []
        for j in range(num_data):
            if adjacency[i][j] == 0:
                neg_distances.append(pdist_matrix[i][j])
            if adjacency[i][j] > 0.0 and i != j:
                pos_distances.append(pdist_matrix[i][j])

        # if there are no positive pairs, distance is 0
        if len(pos_distances) == 0:
            pos_distances.append(0)

        # Sort by distance.
        neg_distances.sort()
        min_neg_distance = neg_distances[0]
        pos_distances.sort(reverse=True)
        max_pos_distance = pos_distances[0]

        if soft:
            loss_np += np.log1p(np.exp(max_pos_distance - min_neg_distance))
        else:
            loss_np += np.maximum(0.0,
                                  max_pos_distance - min_neg_distance + margin)

    loss_np /= num_data
    return loss_np


def test_triplet_loss():
    num_inputs = 10
    # y_true: labels
    y_true = tf.random.uniform((num_inputs, ), 0, 10, dtype=tf.int32)
    # y_preds: embedding
    y_preds = tf.random.uniform((num_inputs, 20), 0, 1)
    tpl = TripletLoss()
    # y_true, y_preds
    loss = tpl(y_true, y_preds)
    assert loss > 0.9


def test_triplet_loss_easy():
    num_inputs = 10
    # y_true: labels
    y_true = tf.random.uniform((num_inputs, ), 0, 3, dtype=tf.int32)
    # y_preds: embedding
    y_preds = tf.random.uniform((num_inputs, 16), 0, 1)
    tpl = TripletLoss(positive_mining_strategy='easy',
                      negative_mining_strategy='easy')
    # y_true, y_preds
    loss = tpl(y_true, y_preds)
    assert loss > 0


def test_triplet_loss_semi_hard():
    num_inputs = 10
    # y_true: labels
    y_true = tf.random.uniform((num_inputs, ), 0, 3, dtype=tf.int32)
    # y_preds: embedding
    y_preds = tf.random.uniform((num_inputs, 16), 0, 1)
    tpl = TripletLoss(positive_mining_strategy='easy',
                      negative_mining_strategy='semi-hard')
    # y_true, y_preds
    loss = tpl(y_true, y_preds)
    assert loss


def test_triplet_loss_hard():
    num_inputs = 10
    # y_true: labels
    y_true = tf.random.uniform((num_inputs, ), 0, 3, dtype=tf.int32)
    # y_preds: embedding
    y_preds = tf.random.uniform((num_inputs, 16), 0, 1)
    tpl = TripletLoss(positive_mining_strategy='hard',
                      negative_mining_strategy='hard')
    # y_true, y_preds
    loss = tpl(y_true, y_preds)
    assert loss


# [pn loss]
def test_pn_loss_serialization():
    loss = PNLoss()
    config = loss.get_config()
    print(config)
    loss2 = PNLoss.from_config(config)
    assert loss.name == loss2.name
    assert loss.distance == loss2.distance


def test_np_loss():
    num_inputs = 10
    # y_true: labels
    y_true = tf.random.uniform((num_inputs, ), 0, 10, dtype=tf.int32)
    # y_preds: embedding
    y_preds = tf.random.uniform((num_inputs, 20), 0, 1)
    pnl = PNLoss()
    # y_true, y_preds
    loss = pnl(y_true, y_preds)
    assert loss > 0.9


# [soft neasrest neighbor loss]
def test_softnn_loss_serialization():
    loss = SoftNearestNeighborLoss(distance="cosine", temperature=50)
    config = loss.get_config()
    loss2 = SoftNearestNeighborLoss.from_config(config)
    assert loss.name == loss2.name
    assert loss.distance == loss2.distance
    assert loss.temperature == loss2.temperature


def softnn_util(y_true, x, temperature = 1):
    """
    A simple loop based implementation of soft
    nearest neighbor loss to test the code.
    https://arxiv.org/pdf/1902.01889.pdf
    """

    y_true = y_true.numpy()
    x = x.numpy()
    batch_size = y_true.shape[0]
    loss = 0
    eps = 1e-9
    for i in range(batch_size):
        numerator = 0
        denominator = 0
        for j in range(batch_size):
            if i == j: continue
            if y_true[i] == y_true[j]:
                numerator += np.exp(-1 * 
                    np.sum(np.square(x[i] - x[j])) / temperature)
            denominator += np.exp(-1 * 
                    np.sum(np.square(x[i] - x[j])) / temperature)
        if numerator == 0: continue
        loss += np.log(numerator/denominator)
    return -loss/batch_size


def test_softnn_loss():
    num_inputs = 10
    n_classes = 10
    # y_true: labels
    y_true = tf.random.uniform((num_inputs,), 0, n_classes, dtype=tf.int32)
    # x: embeddings
    x = tf.random.uniform((num_inputs, 20), 0, 1)

    temperature = np.random.uniform(0.1, 50)
    softnn = SoftNearestNeighborLoss(temperature=temperature)
    loss = softnn(y_true, x)
    loss_check = softnn_util(y_true, x, temperature)
    loss_diff = loss.numpy() - loss_check
    assert np.abs(loss_diff) < 1e-3


def test_xbm_loss():
    batch_size = 6
    embed_dim = 16

    embeddings1 = tf.random.uniform(shape=[batch_size, embed_dim])
    labels1 = tf.constant(
        [
            [1],
            [1],
            [2],
            [2],
            [3],
            [3],
        ],
        dtype=tf.int32
    )

    embeddings2 = tf.random.uniform(shape=[batch_size, embed_dim])
    labels2 = tf.constant(
        [
            [4],
            [4],
            [5],
            [5],
            [6],
            [6],
        ],
        dtype=tf.int32
    )

    distance = "cosine"
    loss = MultiSimilarityLoss(distance=distance)
    loss_nowarm = XBM(loss, memory_size=12, warmup_steps=0)

    # test enqueue
    loss_nowarm(labels1, embeddings1)
    assert loss_nowarm._y_pred_memory.numpy().shape == (batch_size, embed_dim)
    tf.assert_equal(loss_nowarm._y_true_memory, labels1)

    loss_nowarm(labels2, embeddings2)
    assert loss_nowarm._y_pred_memory.numpy().shape == (2*batch_size, embed_dim)
    tf.assert_equal(loss_nowarm._y_true_memory, tf.concat([labels2, labels1], axis=0))

    # test dequeue
    loss_nowarm(labels2, embeddings2)
    assert loss_nowarm._y_pred_memory.numpy().shape == (2 * batch_size, embed_dim)
    tf.assert_equal(loss_nowarm._y_true_memory, tf.concat([labels2, labels2], axis=0))

    # test warmup
    loss_warm = XBM(loss, memory_size=12, warmup_steps=1)

    loss_warm(labels1, embeddings1)
    assert loss_warm._y_pred_memory.numpy().shape == (0, embed_dim)
    tf.assert_equal(loss_warm._y_true_memory, tf.constant([[]], dtype=tf.int32))

    loss_warm(labels2, embeddings2)
    assert loss_warm._y_pred_memory.numpy().shape == (batch_size, embed_dim)
    tf.assert_equal(loss_warm._y_true_memory, labels2)



# Copyright 2021 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""VicReg Loss.
    VICReg: Variance-Invariance-Covariance Regularization for Self-Supervised Learning
    https://arxiv.org/abs/2105.04906
"""
from typing import Any, Callable, Dict, Optional

import tensorflow as tf

from tensorflow_similarity.types import FloatTensor


@tf.keras.utils.register_keras_serializable(package="Similarity")
class VicReg(tf.keras.losses.Loss):
    """VicReg Loss.

    [VICReg: Variance-Invariance-Covariance Regularization for Self-Supervised Learning](https://arxiv.org/abs/2105.04906)
    """  # noqa

    def __init__(
        self,
        std_const: float = 1e-4,
        lambda_: float = 25,
        mu: float = 25,
        nu: float = 1,
        reduction: Callable = tf.keras.losses.Reduction.NONE,
        name: Optional[str] = None,
        **kwargs
    ):
        super().__init__(reduction=reduction, name=name, **kwargs)
        self.lambda_ = lambda_
        self.mu = mu
        self.nu = nu
        self.std_const = std_const
        self.reduction = reduction

    def call(self, za: FloatTensor, zb: FloatTensor) -> FloatTensor:
        """Compute the lost.
        Args:
            za: Embedding A
            zb: Embedding B
        Returns:
            loss
        """
        # compute the diagonal
        batch_size = tf.shape(za)[0]

        # distance loss to measure similarity between representations
        sim_loss = tf.keras.losses.MeanSquaredError(reduction=self.reduction)(
            za, zb
        )
        sim_loss = tf.keras.losses.MeanSquaredError(reduction="none")(za, zb)

        za = self.mean_center_columns(za)
        zb = self.mean_center_columns(zb)

        # std loss to maximize variance(information)
        std_za = tf.sqrt(tf.math.reduce_variance(za, 0) + self.std_const)
        std_zb = tf.sqrt(tf.math.reduce_variance(zb, 0) + self.std_const)

        std_loss_za = tf.reduce_mean(tf.math.maximum(0.0, 1 - std_za))
        std_loss_zb = tf.reduce_mean(tf.math.maximum(0.0, 1 - std_zb))

        std_loss = std_loss_za / 2 + std_loss_zb / 2

        off_diag_ca = self.cov_loss_each(za, batch_size)
        off_diag_cb = self.cov_loss_each(zb, batch_size)

        # covariance loss(1d tensor) for redundancy reduction
        cov_loss = off_diag_ca + off_diag_cb

        loss: FloatTensor = (
            self.lambda_ * sim_loss + self.mu * std_loss + self.nu * cov_loss
        )

        return loss

    def get_config(self) -> Dict[str, Any]:
        config = {
            "std_const": self.std_const,
            "lambda_": self.lambda_,
            "mu": self.mu,
            "nu": self.nu,
        }
        base_config = super().get_config()
        return {**base_config, **config}

    def off_diagonal(self, x: FloatTensor) -> FloatTensor:
        n = tf.shape(x)[0]
        flattened = tf.reshape(x, [-1])[:-1]
        off_diagonals = tf.reshape(flattened, (n - 1, n + 1))[:, 1:]
        off_diag: FloatTensor = tf.reshape(off_diagonals, [-1])
        return off_diag

    def cov_loss_each(self, z, batch_size):
        # cross-correlation matrix axa
        c = tf.matmul(z, z, transpose_a=True)
        c = c / tf.cast(batch_size - 1, dtype="float32")

        num_features = tf.shape(c)[0]

        off_diag_c = self.off_diagonal(c)
        off_diag_c = tf.math.pow(off_diag_c, 2)

        off_diag_c = tf.math.reduce_sum(off_diag_c) / tf.cast(
            num_features, tf.float32
        )

        return off_diag_c

    def mean_center_columns(self, x: FloatTensor) -> FloatTensor:
        col_mean = tf.math.reduce_mean(x, axis=0)

        norm_col: FloatTensor = x - col_mean
        return norm_col



import functools

import tensorflow as tf

from tensorflow_similarity.types import Tensor
from tensorflow_similarity.augmenters.augmentation_utils.random_apply import (
    random_apply,
)


def color_jitter(
    image: Tensor,
    strength: float = 1.0,
    brightness_multiplier=0.8,
    contrast_multiplier=0.8,
    saturation_multiplier=0.8,
    hue_multiplier=0.2,
    random_order: bool = True,
    impl: str = "multiplicative",
) -> Tensor:
    """Distorts the color of the image.

    Args:
      image: The input image tensor.
      strength: the floating number for the strength of the color augmentation.
      random_order: A bool, specifying whether to randomize the jittering order.
      impl: 'additive' or 'multiplicative'.  Whether to use simclrv1 or simclrv2's
          version of random brightness.

    Returns:
      The distorted image tensor.
    """
    brightness = brightness_multiplier * strength
    contrast = contrast_multiplier * strength
    saturation = saturation_multiplier * strength
    hue = hue_multiplier * strength

    if random_order:
        return color_jitter_rand(
            image, brightness, contrast, saturation, hue, impl=impl
        )
    else:
        return color_jitter_nonrand(
            image, brightness, contrast, saturation, hue, impl=impl
        )


def color_jitter_nonrand(
    image: Tensor,
    brightness: float = 0,
    contrast: float = 0,
    saturation: float = 0,
    hue: float = 0,
    impl: str = "multiplicative",
) -> Tensor:
    """Distorts the color of the image (jittering order is fixed).

    Args:
      image: The input image tensor.
      brightness: A float, specifying the brightness for color jitter.
      contrast: A float, specifying the contrast for color jitter.
      saturation: A float, specifying the saturation for color jitter.
      hue: A float, specifying the hue for color jitter.
      impl: 'additive' or 'multiplicative'.  Whether to use simclrv1 or simclrv2's
          version of random brightness.

    Returns:
      The distorted image tensor.
    """
    with tf.name_scope("distort_color"):

        def apply_transform(
            i: int,
            x: Tensor,
            brightness: float,
            contrast: float,
            saturation: float,
            hue: float,
        ) -> Tensor:
            """Apply the i-th transformation."""
            if brightness != 0 and i == 0:
                x = random_brightness(x, max_delta=brightness, impl=impl)
            elif contrast != 0 and i == 1:
                x = tf.image.random_contrast(
                    x, lower=1 - contrast, upper=1 + contrast
                )
            elif saturation != 0 and i == 2:
                x = tf.image.random_saturation(
                    x, lower=1 - saturation, upper=1 + saturation
                )
            elif hue != 0:
                x = tf.image.random_hue(x, max_delta=hue)
            return x

        for i in range(4):
            image = apply_transform(
                i, image, brightness, contrast, saturation, hue
            )
            image = tf.clip_by_value(image, 0.0, 1.0)
        return image


def color_jitter_rand(
    image: Tensor,
    brightness: float = 0,
    contrast: float = 0,
    saturation: float = 0,
    hue: float = 0,
    impl: str = "multiplicative",
) -> Tensor:
    """Distorts the color of the image (jittering order is random).

    Args:
      image: The input image tensor.
      brightness: A float, specifying the brightness for color jitter.
      contrast: A float, specifying the contrast for color jitter.
      saturation: A float, specifying the saturation for color jitter.
      hue: A float, specifying the hue for color jitter.
      impl: 'additive' or 'multiplicative'.  Whether to use simclrv1 or simclrv2's
          version of random brightness.

    Returns:
      The distorted image tensor.
    """
    with tf.name_scope("distort_color"):

        def apply_transform(i, x):
            """Apply the i-th transformation."""

            def brightness_foo():
                if brightness == 0:
                    return x
                else:
                    return random_brightness(x, max_delta=brightness, impl=impl)

            def contrast_foo():
                if contrast == 0:
                    return x
                else:
                    return tf.image.random_contrast(
                        x, lower=1 - contrast, upper=1 + contrast
                    )

            def saturation_foo():
                if saturation == 0:
                    return x
                else:
                    return tf.image.random_saturation(
                        x, lower=1 - saturation, upper=1 + saturation
                    )

            def hue_foo():
                if hue == 0:
                    return x
                else:
                    return tf.image.random_hue(x, max_delta=hue)

            x = tf.cond(
                tf.less(i, 2),
                lambda: tf.cond(tf.less(i, 1), brightness_foo, contrast_foo),
                lambda: tf.cond(tf.less(i, 3), saturation_foo, hue_foo),
            )
            return x

        perm = tf.random.shuffle(tf.range(4))
        for i in range(4):
            image = apply_transform(perm[i], image)
            image = tf.clip_by_value(image, 0.0, 1.0)

        return image


def to_grayscale(image: Tensor, keep_channels: bool = True) -> Tensor:
    image = tf.image.rgb_to_grayscale(image)
    if keep_channels:
        image = tf.tile(image, [1, 1, 3])
    return image


def random_brightness(
    image: Tensor, max_delta: float, impl: str = "multiplicative"
) -> Tensor:
    """A multiplicative vs additive change of brightness.

    Args:
      image: The input image tensor.
      max_delta: Will randomly apply a brightness between [-max_delta, max_delta).
      impl: 'additive' or 'multiplicative'.  Whether to use simclrv1 or simclrv2's
          version of random brightness.

    Returns:
      The brightned image tensor.
    """
    if impl == "multiplicative":
        factor = tf.random.uniform(
            [], tf.maximum(1.0 - max_delta, 0), 1.0 + max_delta
        )
        image = image * factor
    elif impl == "additive":
        image = tf.image.random_brightness(image, max_delta=max_delta)
    else:
        raise ValueError("Unknown impl {} for random brightness.".format(impl))
    return image


def random_color_jitter(
    image: Tensor,
    p_execute=1.0,
    p_jitter: float = 0.8,
    brightness_multiplier=0.8,
    contrast_multiplier=0.8,
    saturation_multiplier=0.8,
    hue_multiplier=0.2,
    p_grey: float = 0.2,
    strength: float = 1.0,
    impl: str = "multiplicative",
) -> Tensor:
    def _transform(image: Tensor) -> Tensor:
        color_jitter_t = functools.partial(
            color_jitter,
            strength=strength,
            brightness_multiplier=brightness_multiplier,
            contrast_multiplier=contrast_multiplier,
            saturation_multiplier=saturation_multiplier,
            hue_multiplier=hue_multiplier,
            impl=impl,
        )
        image = random_apply(color_jitter_t, p=p_jitter, x=image)
        return random_apply(to_grayscale, p=p_grey, x=image)

    return random_apply(_transform, p=p_execute, x=image)



from absl.testing import parameterized
import tensorflow as tf

from tensorflow_similarity.algebra import masked_min, masked_max
from tensorflow_similarity.algebra import build_masks


class LossTest(tf.test.TestCase, parameterized.TestCase):
    def test_mask(self):
        batch_size = 16
        labels = tf.random.uniform((batch_size, 1), 0, 10, dtype=tf.int32)
        positive_mask, negative_mask = build_masks(labels, labels, batch_size)
        self.assertFalse(positive_mask[0][0])
        self.assertFalse(positive_mask[5][5])

        combined = tf.math.logical_or(negative_mask, positive_mask)
        self.assertFalse(combined[0][0])
        for i in range(1, batch_size):
            self.assertTrue(combined[0][i])
            self.assertTrue(combined[i][0])

    @parameterized.named_parameters(
        {
            "testcase_name": "_rank1_labels",
            "query_labels": [1, 2, 0],
            "key_labels": [1, 2, 3, 0],
            "batch_size": 3,
        },
        {
            "testcase_name": "_rank2_labels",
            "query_labels": [[1], [2], [0]],
            "key_labels": [[1], [2], [3], [0]],
            "batch_size": 3,
        },
    )
    def test_mask_non_square(self, query_labels, key_labels, batch_size):
        query_labels = tf.constant(
            query_labels,
            dtype=tf.int32,
        )

        key_labels = tf.constant(
            key_labels,
            dtype=tf.int32,
        )

        positive_mask_nodiag, negative_mask_nodiag = build_masks(
            query_labels, key_labels, batch_size
        )
        positive_mask_wdiag, negative_mask_wdiag = build_masks(
            query_labels,
            key_labels,
            batch_size,
            remove_diagonal=False,
        )

        target_positive_mask_nodiag = tf.constant(
            [
                [False, False, False, False],
                [False, False, False, False],
                [False, False, False, True],
            ]
        )
        self.assertAllEqual(positive_mask_nodiag, target_positive_mask_nodiag)

        target_positive_mask_wdiag = tf.constant(
            [
                [True, False, False, False],
                [False, True, False, False],
                [False, False, False, True],
            ]
        )

        self.assertAllEqual(positive_mask_wdiag, target_positive_mask_wdiag)
        self.assertAllEqual(negative_mask_wdiag, negative_mask_nodiag)

    def test_masked_max(self):
        distances = tf.constant(
            [[1.0, 2.0, 3.0, 0.0], [4.0, 2.0, 1.0, 0.0]], dtype=tf.float32
        )
        mask = tf.constant([[0, 1, 1, 1], [0, 1, 1, 1]], dtype=tf.float32)
        vals, arg_max = masked_max(distances, mask)

        self.assertEqual(vals.shape, (2, 1))
        self.assertEqual(arg_max.shape, (2,))
        self.assertEqual(vals[0], [3.0])
        self.assertEqual(vals[1], [2.0])
        self.assertEqual(arg_max[0], [2])
        self.assertEqual(arg_max[1], [1])

    def test_arg_max_all_unmasked_vals_lt_zero(self):
        # Ensure reduce_max works when all unmasked vals < 0.0.
        distances = tf.constant(
            [[-7.0, -2.0, 7.0, -9.0], [-7.0, 1e-05, 7.0, -9.0]],
            dtype=tf.float32,
        )
        mask = tf.constant([[0, 0, 0, 1], [0, 1, 0, 0]], dtype=tf.float32)
        vals, arg_max = masked_max(distances, mask)

        self.assertEqual(vals.shape, (2, 1))
        self.assertEqual(arg_max.shape, (2,))
        self.assertEqual(vals[0], [-9.0])
        self.assertEqual(vals[1], [1e-05])
        self.assertEqual(arg_max[0], [3])
        self.assertEqual(arg_max[1], [1])

    def test_masked_min(self):
        distances = tf.constant(
            [[1.0, 2.0, 3.0, 0.0], [4.0, 0.0, 1.0, 0.0]], dtype=tf.float32
        )
        mask = tf.constant([[0, 1, 1, 0], [1, 0, 1, 0]], dtype=tf.float32)
        vals, arg_min = masked_min(distances, mask)

        self.assertEqual(vals.shape, (2, 1))
        self.assertEqual(arg_min.shape, (2,))
        self.assertEqual(vals[0], [2.0])
        self.assertEqual(vals[1], [1.0])
        self.assertEqual(arg_min[0], [1])
        self.assertEqual(arg_min[1], [2])

    def test_arg_min_all_unmasked_vals_gt_zero(self):
        # Ensure reduce_max works when all unmasked vals > 0.0.
        distances = tf.constant(
            [[-7.0, -2.0, 7.0, -9.0], [-1e-06, -2.0, 7.0, -9.0]],
            dtype=tf.float32,
        )
        mask = tf.constant([[0, 0, 1, 0], [1, 0, 0, 0]], dtype=tf.float32)
        vals, arg_min = masked_min(distances, mask)

        self.assertEqual(vals.shape, (2, 1))
        self.assertEqual(arg_min.shape, (2,))
        self.assertEqual(vals[0], [7.0])
        self.assertEqual(vals[1], [-1e-06])
        self.assertEqual(arg_min[0], [2])
        self.assertEqual(arg_min[1], [0])


if __name__ == "__main__":
    tf.test.main()



# Copyright 2020 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""SimSiam Loss.
    Exploring Simple Siamese Representation Learning
    https://bit.ly/3LxsWdj
"""
import math
from typing import Any, Callable, Dict, Optional

import tensorflow as tf

from tensorflow_similarity.types import FloatTensor


def negative_cosine_sim(sim: FloatTensor) -> FloatTensor:
    loss: FloatTensor = tf.constant([-1.0]) * sim
    return loss


def cosine_distance(sim: FloatTensor) -> FloatTensor:
    loss: FloatTensor = tf.constant([1.0]) - sim
    return loss


def angular_distance(sim: FloatTensor) -> FloatTensor:
    loss: FloatTensor = tf.math.acos(sim) / tf.constant(math.pi)
    return loss


@tf.keras.utils.register_keras_serializable(package="Similarity")
class SimSiamLoss(tf.keras.losses.Loss):
    """SimSiam Loss.

    Introduced in: [Exploring Simple Siamese Representation Learning](https://bit.ly/3LxsWdj)
    """

    def __init__(
        self,
        projection_type: str = "negative_cosine_sim",
        margin: float = 0.001,
        reduction: Callable = tf.keras.losses.Reduction.AUTO,
        name: Optional[str] = None,
        **kwargs,
    ):
        """Create the SimSiam Loss.

        Args:
          projection_type: Projects results into a metric space to allow KNN
          search.
            negative_cosine_sim: -1.0 * cosine similarity.
            cosine_distance: 1.0 - cosine similarity.
            angular_distance: 1.0 - angular similarity.
          margin: Offset to prevent a distance of 0.
          reduction: (Optional) Type of `tf.keras.losses.Reduction` to apply to
            loss. Default value is `AUTO`.
          name: (Optional) name for the loss.
          **kwargs: The keyword arguments that are passed on to `fn`.
        """
        super().__init__(reduction=reduction, name=name, **kwargs)
        self.projection_type = projection_type
        self.margin = margin

        if self.projection_type == "negative_cosine_sim":
            self._projection = negative_cosine_sim
        elif self.projection_type == "cosine_distance":
            self._projection = cosine_distance
        elif self.projection_type == "angular_distance":
            self._projection = angular_distance
        else:
            raise ValueError(f"{self.projection_type} is not supported.")

    def call(
        self, projector: FloatTensor, predictor: FloatTensor
    ) -> FloatTensor:
        """Compute the loss.

        Notes:
        - Stopping the gradient is critical according to the paper for convergence.

        Args:
            projector: Projector outputs
            predictor: Predictor outputs

        Returns:
            The per example distance between projector_i and predictor_i.
        """

        projector = tf.math.l2_normalize(projector, axis=1)
        predictor = tf.math.l2_normalize(predictor, axis=1)

        # 2D tensor
        vals = predictor * projector
        # 1D tensor
        cosine_simlarity = tf.reduce_sum(vals, axis=1)

        per_example_projection = self._projection(cosine_simlarity)

        # 1D tensor
        loss: FloatTensor = (
            tf.math.multiply(per_example_projection * 0.5) + self.margin
        )

        return loss

    def get_config(self) -> Dict[str, Any]:
        config = {
            "projection_type": self.projection_type,
            "margin": self.margin,
        }
        base_config = super().get_config()
        return {**base_config, **config}



# Copyright 2021 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from typing import Callable, Optional

import tensorflow as tf
from absl import logging


def TFRecordDatasetSampler(
    shard_path: str,
    deserialization_fn: Callable,
    example_per_class: int = 2,
    batch_size: int = 32,
    shards_per_cycle: int = None,
    compression: Optional[str] = None,
    parallelism: int = tf.data.AUTOTUNE,
    async_cycle: bool = False,
    prefetch_size: Optional[int] = None,
    shard_suffix: str = "*.tfrec",
    num_repeat: int = -1,
) -> tf.data.Dataset:
    """Create a [TFRecordDataset](https://www.tensorflow.org/api_docs/python/tf/data/TFRecordDataset) based sampler.

    This sampler should be used when using a TFDataset or have a large
    dataset that needs to be stored on file.

    **WARNING**: This sampler requires that each TF Record file contain
    contiguous blocks of classes where the size of each block is a multiple
    of example_per_class.

    For example, if example_per_class is 2, and we have two TF Record files, then we
    would expect the classes to grouped something like the following.

    01.tf_rec: [0, 0, 0, 0, 1, 1, 1, 1]
    02.tf_rec: [2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3]

    Overall the way we make the sampling process is by using the
    [tf.dataset.interleaves](https://www.tensorflow.org/api_docs/python/tf/data/Dataset#interleave)
    in a non orthodox way: we use its `block_length` to control the
    number of example per class. Relying on pure tf.data ops also
    ensure good compatibility with distribution strategy.


    Args:
        shard_path: Directory where the shards are stored.

        deserialization_fn: Function used to deserialize the tfRecord and
        construct a valid example.

        example_per_class: Number of example per class in each batch.
        Defaults to 2.

        batch_size: How many examples in each batch. The number of class in
        the batch will be `batch_size // example_per_class`.
        Defaults to 32.

        shards_per_cycle: How many shards to use concurrently per cycle.
        Default is None which is all of them. Can cause segv if too many
        shards.

        compression: Which compression was used when creating the dataset.
        `{None, "ZLIB", or "GZIP"}` as specified in
        [TFRecordDataset documentation](https://www.tensorflow.org/api_docs/python/tf/data/TFRecordDataset)
        Defaults to None.

        parallelism: How many parallel calls to do. If not set, will let
        TensorFlow decide by using `tf.data.AUTOTUNE` (-1).

        async_cycle: If True, create a threadpool of size `batch_size //
        example_per_class` and fetch inputs from the cycle shards
        asynchronously; however, in practice, the default single thread setting
        is faster. We only recommend setting this to True if it is absolutely
        necessary.

        prefetch_size: How many batch to precache. Defaults to 10.

        shard_suffix: Glog pattern used to collect the shard files list.
        Defaults to "*.tfrec".

        num_repeat: How many times to repeat the dataset. Defaults to -1 (infinite).

    Returns:
        A `TF.data.dataset` ready to be consumed by the model.
    """
    shards_list = [
        i.decode()
        for i in tf.io.matching_files(os.path.join(shard_path, shard_suffix))
        .numpy()
        .tolist()
    ]
    logging.debug(f"found {shards_list}")
    total_shards = len(shards_list)
    logging.info(f"found {total_shards} shards")

    if not prefetch_size:
        prefetch_size = 10

    # how many shard to iterate over in parallels.
    cycle_length = shards_per_cycle if shards_per_cycle else total_shards
    # how many threads to use when fetching inputs from the cycle shards
    num_parallel_calls = cycle_length if async_cycle else 1

    # shuffle the shard order
    ds = tf.data.Dataset.from_tensor_slices(shards_list)

    # shuffle shard order
    ds = ds.shuffle(total_shards)

    # deterministic must be True here to ensure that we get the correct
    # number of examples per class per batch.
    ds = ds.interleave(
        lambda x: tf.data.TFRecordDataset(
            x,
            compression_type=compression,
        ),  # noqa
        cycle_length=cycle_length,
        block_length=example_per_class,
        num_parallel_calls=num_parallel_calls,
        deterministic=True,
    )
    ds = ds.map(deserialization_fn, num_parallel_calls=parallelism)
    ds = ds.repeat(count=num_repeat)
    ds = ds.batch(batch_size)
    ds = ds.prefetch(prefetch_size)
    return ds



import os
from typing import List, Optional, Any

from functools import partial
import tensorflow as tf

from tensorflow_similarity.types import Tensor
from tensorflow_similarity.augmenters import Augmenter
from .augmentation_utils.cropping import random_resized_crop
from .augmentation_utils.flip import random_random_flip_left_right
from .augmentation_utils.color_jitter import random_color_jitter
from .augmentation_utils.blur import random_blur
from .augmentation_utils.solarize import random_solarize


@tf.function
def augment_barlow(
    image: Tensor,
    height: int,
    width: int,
    flip_probability=0.5,
    brightness_multiplier=0.8,
    contrast_multiplier=0.6,
    saturation_multiplier=0.6,
    hue_multiplier=0.2,
    jitter_probability=0.8,
    greyscale_probability=0.2,
    blur_probability=0.2,
    blur_min_sigma=0,
    blur_max_sigma=1,
    solarize_probability=0.2,
    solarize_pixel_min=0,
    solarize_pixel_max=255,
    solarize_thresh=10,
):
    image = tf.cast(image, dtype="float32")
    image = random_resized_crop(image, height, width)
    image = random_random_flip_left_right(image, p=flip_probability)
    image = random_color_jitter(
        image,
        p_jitter=jitter_probability,
        p_grey=greyscale_probability,
        strength=1.0,
        brightness_multiplier=brightness_multiplier,
        contrast_multiplier=contrast_multiplier,
        saturation_multiplier=saturation_multiplier,
        hue_multiplier=hue_multiplier,
        impl="additive",
    )
    image = random_blur(
        image=image,
        height=height,
        width=width,
        p=blur_probability,
        min_sigma=blur_min_sigma,
        max_sigma=blur_max_sigma,
    )
    image = random_solarize(
        image,
        thresh=solarize_thresh,
        p=solarize_probability,
        pixel_min=solarize_pixel_min,
        pixel_max=solarize_pixel_max,
    )
    image = tf.clip_by_value(image, 0, 1)

    return image


class BarlowAugmenter(Augmenter):
    def __init__(
        self,
        width: int,
        height: int,
        flip_probability=0.5,
        brightness_multiplier=0.8,
        contrast_multiplier=0.6,
        saturation_multiplier=0.6,
        hue_multiplier=0.2,
        jitter_probability=0.8,
        greyscale_probability=0.2,
        blur_probability=0.2,
        blur_min_sigma=0,
        blur_max_sigma=1,
        solarize_probability=0.2,
        solarize_pixel_min=0,
        solarize_pixel_max=255,
        solarize_thresh=10,
        num_cpu: Optional[int] = os.cpu_count(),
    ):
        super().__init__()
        self.num_cpu = num_cpu
        self.width = width
        self.height = height
        self.flip_probability = flip_probability
        self.brightness_multiplier = brightness_multiplier
        self.contrast_multiplier = contrast_multiplier
        self.saturation_multiplier = saturation_multiplier
        self.hue_multiplier = hue_multiplier
        self.jitter_probability = jitter_probability
        self.greyscale_probability = greyscale_probability
        self.blur_probability = blur_probability
        self.blur_min_sigma = blur_min_sigma
        self.blur_max_sigma = blur_max_sigma
        self.solarize_probability = solarize_probability
        self.solarize_pixel_min = solarize_pixel_min
        self.solarize_pixel_max = solarize_pixel_max
        self.solarize_thresh = solarize_thresh

    @tf.function
    def augment(
        self,
        x: Any,
        y: Any = None,
        num_augmentations_per_example: int = 2,
        is_warmup: bool = True,
    ) -> List[Any]:

        with tf.device("/cpu:0"):
            if y is None:
                y = tf.constant([0])
            inputs = tf.stack(x)
            inputs = tf.cast(inputs, dtype="float32")

            views = []

            augment_fn = partial(
                augment_barlow,
                # image=img,
                height=self.height,
                width=self.width,
                flip_probability=self.flip_probability,
                brightness_multiplier=self.brightness_multiplier,
                contrast_multiplier=self.contrast_multiplier,
                saturation_multiplier=self.saturation_multiplier,
                hue_multiplier=self.hue_multiplier,
                jitter_probability=self.jitter_probability,
                greyscale_probability=self.greyscale_probability,
                blur_probability=self.blur_probability,
                blur_min_sigma=self.blur_min_sigma,
                blur_max_sigma=self.blur_max_sigma,
                solarize_probability=self.solarize_probability,
                solarize_pixel_min=self.solarize_pixel_min,
                solarize_pixel_max=self.solarize_pixel_max,
                solarize_thresh=self.solarize_thresh,
            )
            for _ in range(num_augmentations_per_example):

                view = tf.map_fn(
                    lambda img: augment_fn(image=img),
                    inputs,
                    parallel_iterations=self.num_cpu,
                )
                views.append(view)
        return views

    def __call__(
        self,
        x: Any,
        y: Any = None,
        num_augmentations_per_example: int = 2,
        is_warmup: bool = True,
    ) -> List[Any]:
        return list(self.augment(x, y))



from unittest import mock

import tensorflow as tf

from tensorflow_similarity.visualization import confusion_matrix


def test_confusion_matrix_normalize_true():
    y_true = tf.constant([0, 1, 2, 3, 4, 0, 1, 2, 3, 4])
    y_pred = tf.constant([0, 1, 2, 4, 3, 0, 1, 4, 4, 4])

    _, cm = confusion_matrix(y_pred, y_true, show=False)

    expected_cm = tf.constant(
        [
            [1.0, 0.0, 0.0, 0.0, 0.0],
            [0.0, 1.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, 0.5, 0.0, 0.5],
            [0.0, 0.0, 0.0, 0.0, 1.0],
            [0.0, 0.0, 0.0, 0.5, 0.5],
        ],
        dtype="float32",
    )

    assert tf.reduce_all(tf.math.equal(cm, expected_cm))


def test_confusion_matrix_normalize_false():
    y_true = tf.constant([0, 1, 2, 3, 4, 0, 1, 2, 3, 4])
    y_pred = tf.constant([0, 1, 2, 4, 3, 0, 1, 4, 4, 4])

    _, cm = confusion_matrix(y_pred, y_true, normalize=False, show=False)

    expected_cm = tf.constant(
        [
            [2.0, 0.0, 0.0, 0.0, 0.0],
            [0.0, 2.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, 1.0, 0.0, 1.0],
            [0.0, 0.0, 0.0, 0.0, 2.0],
            [0.0, 0.0, 0.0, 1.0, 1.0],
        ],
        dtype="float32",
    )

    assert tf.reduce_all(tf.math.equal(cm, expected_cm))


def test_confusion_matrix_plot():
    y_true = tf.constant([0, 1, 2, 3, 4, 0, 1, 2, 3, 4])
    y_pred = tf.constant([0, 1, 2, 4, 3, 0, 1, 4, 4, 4])

    ax, _ = confusion_matrix(y_pred, y_true, normalize=True, show=False)

    expected_cell_text = [
        "1.00", "0", "0", "0", "0", "0", "1.00", "0", "0", "0", "0", "0",
        "0.50", "0", "0.50", "0", "0", "0", "0", "1.00", "0", "0", "0", "0.50",
        "0.50"
    ]
    tl = ['−1', '0', '1', '2', '3', '4', '5']

    # Check the cell text
    assert all([a._text == b for a, b in zip(ax.texts, expected_cell_text)])
    # Check the axes text
    assert ax.title._text == 'Confusion matrix'
    assert ax.get_xlabel() == 'Predicted label\naccuracy=0.6000; misclass=0.4000'
    assert ax.get_ylabel() == 'True label'
    # Check the ticks are just integers.
    assert all([a._text == b for a, b in zip(ax.get_xticklabels(), tl)])
    assert all([a._text == b for a, b in zip(ax.get_yticklabels(), tl)])


def test_confusion_matrix_plot_labels():
    y_true = tf.constant([0, 1, 2, 3, 4, 0, 1, 2, 3, 4])
    y_pred = tf.constant([0, 1, 2, 4, 3, 0, 1, 4, 4, 4])
    lbls = ['A', 'B', 'C', 'D', 'E']

    ax, _ = confusion_matrix(
        y_pred,
        y_true,
        labels=lbls,
        show=False
    )

    # Check all tick labels match the provided list
    assert all([a._text == b for a, b in zip(ax.get_xticklabels(), lbls)])
    assert all([a._text == b for a, b in zip(ax.get_yticklabels(), lbls)])


def test_confusion_matrix_title():
    y_true = tf.constant([0, 1, 2, 3, 4, 0, 1, 2, 3, 4])
    y_pred = tf.constant([0, 1, 2, 4, 3, 0, 1, 4, 4, 4])

    ax, _ = confusion_matrix(
        y_pred,
        y_true,
        title='foo',
        show=False
    )

    # Check the main title
    assert ax.title._text == 'foo'


@mock.patch("tensorflow_similarity.visualization.confusion_matrix_viz.plt.show")
@mock.patch(
    "tensorflow_similarity.visualization.confusion_matrix_viz.plt.subplots",
    autospec=True,
)
def test_confusion_matrix_mock(mock_subplots, mock_show):
    ax = mock.Mock()
    fig = mock.Mock()
    mock_subplots.return_value = (fig, ax)

    y_true = tf.constant([0, 1, 2, 3, 4, 0, 1, 2, 3, 4])
    y_pred = tf.constant([0, 1, 2, 4, 3, 0, 1, 4, 4, 4])

    _, cm = confusion_matrix(
        y_pred,
        y_true,
        cmap='Reds',
        show=False
    )
    mock_subplots.assert_called_with(figsize=(8, 6))
    ax.imshow.assert_called_with(
            cm,
            interpolation='nearest',
            cmap="Reds"
    )
    fig.colorbar.assert_called_with(ax.imshow())
    mock_show.assert_not_called()

    _, cm = confusion_matrix(
        y_pred,
        y_true,
        cmap='Reds',
        show=True
    )
    mock_show.assert_called_once()





import tensorflow as tf

from tensorflow_similarity.types import Tensor
from tensorflow_similarity.augmenters.augmentation_utils.random_apply import (
    random_apply,
)


def random_solarize(
    image: Tensor,
    p: float = 0.2,
    pixel_min=0,
    pixel_max=255,
    thresh: int = 10,
) -> Tensor:
    def _transform(image: Tensor) -> Tensor:
        return tf.where(image < thresh, image, pixel_max - image)

    return random_apply(_transform, p=p, x=image)



# Copyright 2020 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Soft Nearest Neighbors Loss.
    FaceNet: A Unified Embedding for Face Recognition and Clustering
    https://arxiv.org/abs/1902.01889
"""
from typing import Any, Callable, Union

import tensorflow as tf

from tensorflow_similarity.algebra import build_masks
from tensorflow_similarity.distances import Distance, distance_canonicalizer
from tensorflow_similarity.types import FloatTensor, IntTensor

from .metric_loss import MetricLoss


def soft_nn_loss(
    query_labels: IntTensor,
    query_embeddings: FloatTensor,
    key_labels: IntTensor,
    key_embeddings: FloatTensor,
    distance: Callable,
    temperature: float,
    remove_diagonal: bool = True,
) -> Any:
    """Computes the soft nearest neighbors loss.

    Args:
        query_labels: labels associated with the query embed.
        query_embeddings: Embedded query examples.
        key_labels: labels associated with the key embed.
        key_embeddings: Embedded key examples.
        distance: Which distance function to use to compute the pairwise.
        temperature: Controls relative importance given
                        to the pair of points.
        remove_diagonal: Bool. If True, will set diagonal to False in positive pair mask

    Returns:
        loss: loss value for the current batch.
    """

    batch_size = tf.size(query_labels)
    eps = 1e-9

    pairwise_dist = distance(query_embeddings, key_embeddings)
    pairwise_dist = pairwise_dist / temperature
    negexpd = tf.math.exp(-pairwise_dist)

    # Mask out diagonal entries
    diag = tf.linalg.diag(tf.ones(batch_size, dtype=tf.bool))
    diag_mask = tf.cast(tf.logical_not(diag), dtype=tf.float32)
    negexpd = tf.math.multiply(negexpd, diag_mask)

    # creating mask to sample same class neighboorhood
    pos_mask, _ = build_masks(
        query_labels,
        key_labels,
        batch_size=batch_size,
        remove_diagonal=remove_diagonal,
    )
    pos_mask = tf.cast(pos_mask, dtype=tf.float32)

    # all class neighborhood
    alcn = tf.reduce_sum(negexpd, axis=1)

    # same class neighborhood
    sacn = tf.reduce_sum(tf.math.multiply(negexpd, pos_mask), axis=1)

    # exclude examples with unique class from loss calculation
    excl = tf.math.not_equal(
        tf.reduce_sum(pos_mask, axis=1), tf.zeros(batch_size)
    )
    excl = tf.cast(excl, tf.float32)

    loss = tf.math.divide(sacn, alcn)
    loss = -tf.multiply(tf.math.log(eps + loss), excl)

    return loss


@tf.keras.utils.register_keras_serializable(package="Similarity")
class SoftNearestNeighborLoss(MetricLoss):
    """Computes the soft nearest neighbors loss in an online fashion.

    Similar to TripletLoss, this loss compares intra- and inter-class
    distances. However, unlike TripletLoss, this loss uses all
    positive and negative examples in the batch.
    See: https://arxiv.org/abs/1902.01889 for the original paper.

    `labels` must be  a 1-D integer `Tensor` of shape (batch_size,).
    It's values represent the classes associated with the examples as
    **integer  values**.

    `embeddings` must be 2-D float `Tensor` of embedding vectors.
    """

    def __init__(
        self,
        distance: Union[Distance, str] = "sql2",
        temperature: float = 1,
        name: str = "SoftNearestNeighborLoss",
        **kwargs
    ):
        """Initializes the SoftNearestNeighborLoss Loss

        Args:
            `distance`: Which distance function to use to compute
                        the pairwise distances between embeddings.
                        Defaults to 'sql2'.

            `temperature`: Alters the value of loss function.
                           Defaults to 1.

            `name`: Loss name. Defaults to SoftNearestNeighborLoss.
        """
        # distance canonicalization
        distance = distance_canonicalizer(distance)
        self.distance = distance
        self.temperature = temperature

        super().__init__(
            fn=soft_nn_loss,
            name=name,
            distance=distance,
            temperature=temperature,
            **kwargs
        )



import math
from unittest import TestCase

import numpy as np
import pytest
import tensorflow as tf

from tensorflow_similarity.layers import (GeneralizedMeanPooling1D,
                                          GeneralizedMeanPooling2D,
                                          MetricEmbedding)


@pytest.fixture
def input_2d_tensor():
    # A (1, 3, 3, 2) Tensor.
    return tf.constant(
        [
            [
                [[0.0, 5.0], [0.0, 5.0], [0.0, 5.0]],
                [[5.0, 10.0], [5.0, 10.0], [5.0, 10.0]],
                [[10.0, 15.0], [10.0, 15.0], [10.0, 15.0]],
            ],
        ]
    )


def test_generalized_mean_pooling_2d(input_2d_tensor):
    result = GeneralizedMeanPooling2D(p=1.0)(input_2d_tensor)
    expected = tf.constant([[5.0, 10.0]])
    np.testing.assert_allclose(result, expected, rtol=1e-06)


def test_generalized_mean_pooling_2d_inf(input_2d_tensor):
    result = GeneralizedMeanPooling2D(p=math.inf, keepdims=True)(
        input_2d_tensor
    )
    expected = tf.constant([[[[10.0, 15.0]]]])
    np.testing.assert_allclose(result, expected, rtol=1e-06)


def test_generalized_mean_pooling_2d_neg_inf(input_2d_tensor):
    result = GeneralizedMeanPooling2D(p=-math.inf, keepdims=True)(
        input_2d_tensor
    )
    expected = tf.constant([[[[0.0, 5.0]]]])
    np.testing.assert_allclose(result, expected, rtol=1e-06)


def test_generalized_mean_pooling_2d_zero(input_2d_tensor):
    result = GeneralizedMeanPooling2D(p=0.0)(input_2d_tensor)
    expected = tf.constant([[3.0412402, 8.041241]])
    np.testing.assert_allclose(result, expected, rtol=1e-06)


def test_generalized_mean_pooling_2d_keepdims(input_2d_tensor):
    result = GeneralizedMeanPooling2D(p=1.0, keepdims=True)(input_2d_tensor)
    expected = tf.constant([[[[5.0, 10.0]]]])
    np.testing.assert_allclose(result, expected, rtol=1e-06)


def test_generalized_mean_pooling_2d_channels_first(input_2d_tensor):
    result = GeneralizedMeanPooling2D(p=1.0, data_format="channels_first")(
        input_2d_tensor
    )
    # With channels first we expect a (1,3) shape.
    expected = tf.constant([[2.5, 7.5, 12.5]])
    np.testing.assert_allclose(result, expected, rtol=1e-06)


def test_generalized_mean_pooling_2d_compute_shape():
    input_shape = tf.constant([1, 2, 3, 4])
    result = GeneralizedMeanPooling2D().compute_output_shape(input_shape)
    expected_shape = tf.constant([1, 4])
    np.testing.assert_allclose(result, expected_shape, rtol=1e-06)


def test_generalized_mean_pooling_2d_compute_shape_keepdims():
    input_shape = tf.constant([1, 2, 3, 4])
    result = GeneralizedMeanPooling2D(keepdims=True).compute_output_shape(
        input_shape
    )
    expected_shape = tf.constant([1, 1, 1, 4])
    np.testing.assert_allclose(result, expected_shape, rtol=1e-06)


def test_generalized_mean_pooling_2d_compute_shape_dataformat():
    input_shape = tf.constant([1, 2, 3, 4])
    result = GeneralizedMeanPooling2D(
        data_format="channels_first"
    ).compute_output_shape(input_shape)
    expected_shape = tf.constant([1, 2])
    np.testing.assert_allclose(result, expected_shape, rtol=1e-06)


def test_generalized_mean_pooling_2d_get_config():
    gem = GeneralizedMeanPooling2D(
        p=3.0, data_format="channels_first", keepdims=True, name="GEM"
    )
    config = gem.get_config()
    expected_config = {
        "p": 3.0,
        "data_format": "channels_first",
        "keepdims": True,
        "name": "GEM",
        "trainable": True,
        "dtype": "float32",
    }
    TestCase().assertDictEqual(expected_config, config)


@pytest.fixture
def input_1d_tensor():
    # A (1, 3, 1) Tensor.
    x = tf.constant([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0]])
    x = tf.reshape(x, [3, 3, 1])
    return x


def test_generalized_mean_pooling_1d(input_1d_tensor):
    result = GeneralizedMeanPooling1D(p=1.0)(input_1d_tensor)
    expected = tf.constant([[2.0], [5.0], [8.0]])
    np.testing.assert_allclose(result, expected, rtol=1e-06)


def test_generalized_mean_pooling_1d_inf(input_1d_tensor):
    result = GeneralizedMeanPooling1D(p=math.inf, keepdims=True)(
        input_1d_tensor
    )
    expected = tf.constant(
        [[[3.0], [3.0], [3.0]], [[6.0], [6.0], [6.0]], [[9.0], [9.0], [9.0]]]
    )
    np.testing.assert_allclose(result, expected, rtol=1e-06)


def test_generalized_mean_pooling_1d_neg_inf(input_1d_tensor):
    result = GeneralizedMeanPooling1D(p=-math.inf, keepdims=True)(
        input_1d_tensor
    )
    expected = tf.constant(
        [[[1.0], [1.0], [1.0]], [[4.0], [4.0], [4.0]], [[7.0], [7.0], [7.0]]]
    )
    np.testing.assert_allclose(result, expected, rtol=1e-06)


def test_generalized_mean_pooling_1d_zero(input_1d_tensor):
    result = GeneralizedMeanPooling1D(p=0.0)(input_1d_tensor)
    expected = tf.constant([[1.8171206], [4.8171206], [7.8171206]])
    np.testing.assert_allclose(result, expected, rtol=1e-06)


def test_generalized_mean_pooling_1d_keepdims(input_1d_tensor):
    result = GeneralizedMeanPooling1D(p=1.0, keepdims=True)(input_1d_tensor)
    expected = tf.constant([[[2.0]], [[5.0]], [[8.0]]])
    np.testing.assert_allclose(result, expected, rtol=1e-06)


def test_generalized_mean_pooling_1d_channels_first(input_1d_tensor):
    input_1d_tensor = tf.reshape(input_1d_tensor, [3, 1, 3])
    result = GeneralizedMeanPooling1D(p=1.0, data_format="channels_first")(
        input_1d_tensor
    )
    expected = tf.constant([[2.0], [5.0], [8.0]])
    np.testing.assert_allclose(result, expected, rtol=1e-06)


def test_generalized_mean_pooling_1d_compute_shape():
    input_shape = tf.constant([1, 2, 3])
    result = GeneralizedMeanPooling1D().compute_output_shape(input_shape)
    expected_shape = tf.constant([1, 3])
    np.testing.assert_allclose(result, expected_shape, rtol=1e-06)


def test_generalized_mean_pooling_1d_compute_shape_keepdims():
    input_shape = tf.constant([1, 2, 3])
    result = GeneralizedMeanPooling1D(keepdims=True).compute_output_shape(
        input_shape
    )
    expected_shape = tf.constant([1, 1, 3])
    np.testing.assert_allclose(result, expected_shape, rtol=1e-06)


def test_generalized_mean_pooling_1d_compute_shape_dataformat():
    input_shape = tf.constant([1, 2, 3])
    result = GeneralizedMeanPooling1D(
        data_format="channels_first"
    ).compute_output_shape(input_shape)
    expected_shape = tf.constant([1, 2])
    np.testing.assert_allclose(result, expected_shape, rtol=1e-06)


def test_generalized_mean_pooling_1d_get_config():
    gem = GeneralizedMeanPooling1D(
        p=3.0, data_format="channels_first", keepdims=True, name="GEM"
    )
    config = gem.get_config()
    expected_config = {
        "p": 3.0,
        "data_format": "channels_first",
        "keepdims": True,
        "name": "GEM",
        "trainable": True,
        "dtype": "float32",
    }
    TestCase().assertDictEqual(expected_config, config)


def test_metric_embedding():
    input_tensor = tf.constant([[4.0, 4.0, 4.0, 4.0], [1.0, 1.0, 1.0, 1.0]])
    me_layer = MetricEmbedding(
        4, kernel_initializer=tf.constant_initializer(1.0)
    )
    result = me_layer(input_tensor)
    expected_result = tf.constant([[0.5, 0.5, 0.5, 0.5], [0.5, 0.5, 0.5, 0.5]])
    np.testing.assert_allclose(result, expected_result, rtol=1e-06)


def test_metric_embedding_get_config():
    me_layer = MetricEmbedding(32)
    config = me_layer.get_config()
    expected_config = {
        "name": "metric_embedding_1",
        "trainable": True,
        "dtype": "float32",
        "units": 32,
        "activation": "linear",
        "use_bias": True,
        "kernel_initializer": {
            "class_name": "GlorotUniform",
            "config": {"seed": None},
        },
        "bias_initializer": {"class_name": "Zeros", "config": {}},
        "kernel_regularizer": None,
        "bias_regularizer": None,
        "activity_regularizer": None,
        "kernel_constraint": None,
        "bias_constraint": None,
    }
    TestCase().assertDictEqual(expected_config, config)



# Copyright 2021 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tensorflow as tf

from tensorflow_similarity.types import FloatTensor

from .classification_metric import ClassificationMetric


class Precision(ClassificationMetric):
    """Calculates the precision of the query classification.

    Computes the precision given the query classification counts.

    $$
    Precision = \frac{\textrm{true_positives}}{\textrm{true_positives} + \textrm{false_positives}}
    $$

    args:
        name: Name associated with a specific metric object, e.g.,
        precision@0.1

    Usage with `tf.similarity.models.SimilarityModel()`:

    ```python
    model.calibrate(x=query_examples,
                    y=query_labels,
                    calibration_metric='precision')
    ```
    """

    def __init__(self, name: str = "precision") -> None:
        super().__init__(
            name=name, canonical_name="precision", increasing=False
        )

    def compute(
        self,
        tp: FloatTensor,
        fp: FloatTensor,
        tn: FloatTensor,
        fn: FloatTensor,
        count: int,
    ) -> FloatTensor:
        """Compute the classification metric.

        The `compute()` method supports computing the metric for a set of
        values, where each value represents the counts at a specific distance
        threshold.

        Args:
            tp: A 1D FloatTensor containing the count of True Positives at each
            distance threshold.

            fp: A 1D FloatTensor containing the count of False Positives at
            each distance threshold.

            tn: A 1D FloatTensor containing the count of True Negatives at each
            distance threshold.

            fn: A 1D FloatTensor containing the count of False Negatives at
            each distance threshold.

            count: The total number of queries

        Returns:
            A 1D FloatTensor containing the metric at each distance threshold.
        """
        p: FloatTensor = tf.math.divide_no_nan(tp, tp + fp)

        # If all queries return empty result sets we have a recall of zero. In
        # this case the precision should be 1.0 (see
        # https://nlp.stanford.edu/IR-book/html/htmledition/evaluation-of-ranked-retrieval-results-1.html#fig:precision-recall).
        # The following accounts for the and sets the first precision value to
        # 1.0 if the first recall and precision are both zero.
        if (tp + fp)[0] == 0.0 and len(p) > 1:
            initial_precision = tf.constant(
                [tf.constant([1.0]), tf.zeros(len(p) - 1)], axis=0
            )
            p = p + initial_precision

        return p
